#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is herby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import logging
import os
import random
import uuid
from unittest.mock import Mock, call

import pytest

import paleomix.node
from paleomix.atomiccmd.command import AtomicCmd
from paleomix.common.utilities import safe_coerce_to_frozenset
from paleomix.node import (
    CmdNodeError,
    CommandNode,
    Node,
    NodeError,
    NodeUnhandledException,
)


def create_test_dir():
    return os.path.dirname(__file__)


def create_test_file(*args):
    return os.path.join(create_test_dir(), "data", *args)


def first(values):
    return random.choice(tuple(values))


def _CommandNodeWrap(**kwargs):
    return CommandNode(command=AtomicCmd("true"), **kwargs)


_NODE_TYPES = (Node, _CommandNodeWrap)


_DESCRIPTION = "My description of a node"
_IN_FILES = frozenset(
    (create_test_file("empty_file_1"), create_test_file("empty_file_2"))
)
_OUT_FILES = frozenset(
    (create_test_file("missing_out_file_1"), create_test_file("missing_out_file_2"))
)
_EXEC_FILES = frozenset(("ls", "sh"))
_AUX_FILES = frozenset(
    (create_test_file("rCRS.fasta"), create_test_file("rCRS.fasta.fai"))
)
_REQUIREMENTS = frozenset((id, str))
_EMPTY_FILE = create_test_file("empty_file_1")


def _build_cmd_mock(
    input_files=_IN_FILES,
    output_files=(),
    executables=(),
    auxiliary_files=(),
    requirements=(),
    optional_temp_files=(),
    return_codes=(0,),
):
    cmd = Mock(
        input_files=frozenset(input_files),
        output_files=frozenset(output_files),
        executables=frozenset(executables),
        auxiliary_files=frozenset(auxiliary_files),
        requirements=frozenset(requirements),
        expected_temp_files=frozenset(map(os.path.basename, output_files)),
        optional_temp_files=frozenset(optional_temp_files),
    )
    cmd.join.return_value = return_codes

    return cmd


###############################################################################
###############################################################################
# Node: Constructor: File sets


_CONSTUCTOR_SINGLE_VALUES = (
    # Single values
    ("input_files", first(_IN_FILES)),
    ("output_files", first(_OUT_FILES)),
    ("executables", first(_EXEC_FILES)),
    ("auxiliary_files", first(_AUX_FILES)),
    # Single value in list
    ("input_files", [first(_IN_FILES)]),
    ("output_files", [first(_OUT_FILES)]),
    ("executables", [first(_EXEC_FILES)]),
    ("auxiliary_files", [first(_AUX_FILES)]),
    # Multiple values in list
    ("input_files", _IN_FILES),
    ("output_files", _OUT_FILES),
    ("executables", _EXEC_FILES),
    ("auxiliary_files", _AUX_FILES),
)


@pytest.mark.parametrize("key, value", _CONSTUCTOR_SINGLE_VALUES)
def test_constructor(key, value):
    defaults = {"input_files": _EMPTY_FILE}
    defaults[key] = value
    node = Node(**defaults)
    expected = safe_coerce_to_frozenset(value)
    assert getattr(node, key) == expected


_CONSTUCTOR_INVALID_VALUES = (
    ("input_files", [id]),
    ("output_files", [-1]),
    ("executables", [{}]),
    ("auxiliary_files", [1.3]),
)


@pytest.mark.parametrize("key, value", _CONSTUCTOR_INVALID_VALUES)
def test_constructor__invalid_values(key, value):
    with pytest.raises(TypeError):
        Node(**{key: value})


###############################################################################
###############################################################################
# Node: Constructor: Requirements


def test_constructor__requirements():
    node = Node(requirements=id)
    assert node.requirements == frozenset([id])
    node = Node(requirements=[id])
    assert node.requirements == frozenset([id])
    node = Node(requirements=[id, str])
    assert node.requirements == frozenset([id, str])


@pytest.mark.parametrize("value", (17, {}, "867-5309"))
def test_constructor__requirements__wrong_type(value):
    with pytest.raises(TypeError):
        Node(requirements=value)


###############################################################################
###############################################################################
# Node: Constructor: Dependencies


def test_constructor__nodes_is_none():
    my_node = Node(dependencies=None)
    assert my_node.dependencies == frozenset()


def test_constructor__single_node():
    sub_node = Node()
    my_node = Node(dependencies=sub_node)
    assert my_node.dependencies == frozenset([sub_node])


def test_constructor__iterable():
    sub_nodes = [Node(), Node()]
    my_node = Node(dependencies=iter(sub_nodes))
    assert my_node.dependencies == frozenset(sub_nodes)


def test_constructor__not_a_node():
    with pytest.raises(TypeError):
        Node(dependencies=(1,))


###############################################################################
###############################################################################
# *Node: Description


@pytest.mark.parametrize("cls", _NODE_TYPES)
def test_constructor__description(cls):
    my_node = cls(description=_DESCRIPTION)
    assert str(my_node) == _DESCRIPTION


@pytest.mark.parametrize("cls", _NODE_TYPES)
def test_constructor__description__default(cls):
    my_node = cls()
    assert str(my_node) == repr(my_node)


@pytest.mark.parametrize("cls", _NODE_TYPES)
@pytest.mark.parametrize("value", (1, {}))
def test_constructor__description__non_string(cls, value):
    with pytest.raises(TypeError):
        cls(description=value)


###############################################################################
###############################################################################
# *Node: Constructor tests: #threads


@pytest.mark.parametrize("cls", _NODE_TYPES)
@pytest.mark.parametrize("nthreads", (1, 3))
def test_constructor__threads(cls, nthreads):
    node = cls(threads=nthreads)
    assert node.threads == nthreads


@pytest.mark.parametrize("cls", _NODE_TYPES)
@pytest.mark.parametrize("nthreads", (-1, 0))
def test_constructor__threads_invalid_range(cls, nthreads):
    with pytest.raises(ValueError):
        cls(threads=nthreads)


@pytest.mark.parametrize("cls", _NODE_TYPES)
@pytest.mark.parametrize("nthreads", ("1", {}, 2.7))
def test_constructor__threads_invalid_type(cls, nthreads):
    with pytest.raises(TypeError):
        cls(threads=nthreads)


###############################################################################
###############################################################################
# Node: Run

_DUMMY_TEMP_ROOT = "/xyz/tmp"
_DUMMY_TEMP = os.path.join(_DUMMY_TEMP_ROOT, "xTMPx")


def test_run__order():
    cfg_mock = Mock(temp_root=_DUMMY_TEMP_ROOT)
    node_mock = Mock()

    node = Node()
    node._create_temp_dir = node_mock._create_temp_dir
    node._create_temp_dir.return_value = _DUMMY_TEMP
    node._setup = node_mock._setup
    node._run = node_mock._run
    node._teardown = node_mock._teardown
    node._remove_temp_dir = node_mock._remove_temp_dir

    node.run(cfg_mock)

    node_mock.mock_calls == [
        call._create_temp_dir(cfg_mock),
        call._setup(cfg_mock, _DUMMY_TEMP),
        call._run(cfg_mock, _DUMMY_TEMP),
        call._teardown(cfg_mock, _DUMMY_TEMP),
        call._remove_temp_dir(_DUMMY_TEMP),
    ]


_EXCEPTIONS = (
    (TypeError("The castle AAARGH!"), NodeUnhandledException),
    (NodeError("He's a very naughty boy!"), NodeError),
)


@pytest.mark.parametrize("key", ("_setup", "_run", "_teardown"))
@pytest.mark.parametrize("exception, expectation", _EXCEPTIONS)
def test_run__exceptions(key, exception, expectation):
    mock = Mock()
    node = Node()
    node._create_temp_dir = mock._create_temp_dir
    node._create_temp_dir.return_value = _DUMMY_TEMP

    setattr(node, key, getattr(mock, key))
    getattr(node, key).side_effect = exception

    cfg_mock = Mock(temp_root=_DUMMY_TEMP_ROOT)
    with pytest.raises(expectation):
        node.run(cfg_mock)

    assert mock.mock_calls == [
        call._create_temp_dir(cfg_mock),
        getattr(call, key)(cfg_mock, _DUMMY_TEMP),
    ]


def test_run__exception__create_temp_dir():
    cfg_mock = Mock(temp_root=_DUMMY_TEMP_ROOT)
    node_mock = Node()
    node_mock._create_temp_dir = Mock()
    node_mock._create_temp_dir.side_effect = OSError()

    with pytest.raises(NodeUnhandledException):
        node_mock.run(cfg_mock)
    assert node_mock._create_temp_dir.mock_calls == [call(cfg_mock)]


def test_run__exception__remove_temp_dir():
    cfg_mock = Mock(temp_root=_DUMMY_TEMP_ROOT)
    mock = Mock()
    node_mock = Node()
    node_mock._create_temp_dir = mock._create_temp_dir
    node_mock._create_temp_dir.return_value = _DUMMY_TEMP
    node_mock._remove_temp_dir = mock._remove_temp_dir
    node_mock._remove_temp_dir.side_effect = OSError()

    with pytest.raises(NodeUnhandledException):
        node_mock.run(cfg_mock)
    assert mock.mock_calls == [
        call._create_temp_dir(cfg_mock),
        call._remove_temp_dir(_DUMMY_TEMP),
    ]


@pytest.mark.parametrize("exception", (NodeError, OSError))
def test_run__error_log__node_error(tmp_path, exception):
    temp = tmp_path / "xTMPx"
    mock = Mock()
    cfg_mock = Mock(temp_root=tmp_path)
    node_mock = Node()
    node_mock._create_temp_dir = mock._create_temp_dir
    node_mock._create_temp_dir.return_value = str(temp)
    node_mock._run = mock._run
    node_mock._run.side_effect = exception("ARGH!")

    temp.mkdir()
    with pytest.raises(NodeError):
        node_mock.run(cfg_mock)
    log_file = tmp_path / "xTMPx" / "pipe.errors"
    assert log_file.exists()

    error_text = log_file.read_text()
    assert "Errors =" in error_text
    assert paleomix.__version__ in error_text

    assert mock.mock_calls == [
        call._create_temp_dir(cfg_mock),
        call._run(cfg_mock, str(temp)),
    ]


###############################################################################
###############################################################################
# Node: _setup / _teardown

_INPUT_FILES_EXIST = (
    {"executables": ("ls", "sh")},
    {"input_files": _IN_FILES},
    {"auxiliary_files": _IN_FILES},
)


@pytest.mark.parametrize("kwargs", _INPUT_FILES_EXIST)
def test__setup__input_files(kwargs):
    Node(**kwargs)._setup(None, None)


_INPUT_FILES_MISSING = (
    {"executables": ("ls", "shxxxx")},
    {"input_files": _OUT_FILES},
    {"auxiliary_files": _OUT_FILES},
)


@pytest.mark.parametrize("kwargs", _INPUT_FILES_MISSING)
def test__setup__input_files_missing(kwargs):
    with pytest.raises(NodeError):
        Node(**kwargs)._setup(None, None)


def test__teardown__output_files():
    Node(input_files=_EMPTY_FILE, output_files=_IN_FILES)._teardown(None, None)


def test__teardown__output_files_missing():
    node = Node(input_files=_EMPTY_FILE, output_files=_OUT_FILES)
    with pytest.raises(NodeError):
        node._teardown(None, None)


###############################################################################
# Node._remove_temp_dir


def test_node_remove_temp_dir__empty_dir(tmp_path, caplog):
    with caplog.at_level(logging.WARNING):
        node = Node()
        node._remove_temp_dir(tmp_path)

        assert not tmp_path.exists()
        assert not caplog.messages


def test_node_remove_temp_dir__extranous_files(tmp_path, caplog):
    tmp_file = tmp_path / str(uuid.uuid4())
    tmp_file.touch()

    with caplog.at_level(logging.WARNING):
        node = Node()
        node._remove_temp_dir(tmp_path)

        assert not tmp_path.exists()
        assert (
            paleomix.node.__name__,
            logging.WARNING,
            "Unexpected file in temporary directory: %r" % (str(tmp_file),),
        ) in caplog.record_tuples


###############################################################################
# Node._collect_files


def test_node_collect_files__empty_folder(tmp_path):
    assert list(Node._collect_files(tmp_path)) == []


def test_node_collect_files__root_files(tmp_path):
    (tmp_path / "foo.txt").touch()
    (tmp_path / "bar.txt").touch()

    assert sorted(Node._collect_files(tmp_path)) == [
        "bar.txt",
        "foo.txt",
    ]


def test_node_collect_files__files_and_folders(tmp_path):
    (tmp_path / "foo1.txt").touch()
    (tmp_path / "bar1").mkdir()
    (tmp_path / "bar1" / "foo2.txt").touch()
    (tmp_path / "bar1" / "bar2").mkdir()
    (tmp_path / "bar1" / "bar2" / "foo3.txt").touch()
    (tmp_path / "bar1" / "bar2" / "foo4.txt").touch()

    assert sorted(Node._collect_files(tmp_path)) == [
        "bar1/bar2/foo3.txt",
        "bar1/bar2/foo4.txt",
        "bar1/foo2.txt",
        "foo1.txt",
    ]


###############################################################################
###############################################################################
# CommandNode: Constructor

_SIMPLE_DEPS = Node()
_SIMPLE_SUBS = Node()
_SIMPLE_CMD_MOCK = Mock(
    input_files=_IN_FILES,
    output_files=_OUT_FILES,
    executables=_EXEC_FILES,
    auxiliary_files=_AUX_FILES,
    requirements=_REQUIREMENTS,
)
_SIMPLE_CMD_NODE = CommandNode(command=_SIMPLE_CMD_MOCK, dependencies=_SIMPLE_DEPS)


def test_commandnode_constructor__input_files():
    assert _SIMPLE_CMD_NODE.input_files == _IN_FILES


def test_commandnode_constructor__output_files():
    assert _SIMPLE_CMD_NODE.output_files == _OUT_FILES


def test_commandnode_constructor__auxiliary_files():
    assert _SIMPLE_CMD_NODE.auxiliary_files == _AUX_FILES


def test_commandnode_constructor__executables():
    assert _SIMPLE_CMD_NODE.executables == _EXEC_FILES


def test_commandnode_constructor__requirements():
    assert _SIMPLE_CMD_NODE.requirements == _REQUIREMENTS


def test_commandnode_constructor__dependencies():
    assert _SIMPLE_CMD_NODE.dependencies == frozenset([_SIMPLE_DEPS])


def test_commandnode_constructor__dependencies__default():
    cmd_mock = CommandNode(command=_SIMPLE_CMD_MOCK)
    assert cmd_mock.dependencies == frozenset()


###############################################################################
###############################################################################
# CommandNode: run


def test_command_node__run():
    cfg_mock = Mock(temp_root=_DUMMY_TEMP_ROOT)
    mock = _build_cmd_mock()

    node_mock = CommandNode(mock)
    node_mock._create_temp_dir = mock._test_node_._create_temp_dir
    node_mock._create_temp_dir.return_value = _DUMMY_TEMP
    node_mock._setup = mock._test_node_._setup
    node_mock._teardown = mock._test_node_._teardown
    node_mock._remove_temp_dir = mock._test_node_._remove_temp_dir

    node_mock.run(cfg_mock)

    assert mock.mock_calls == [
        call._test_node_._create_temp_dir(cfg_mock),
        call._test_node_._setup(cfg_mock, _DUMMY_TEMP),
        call.run(_DUMMY_TEMP),
        call.join(),
        call._test_node_._teardown(cfg_mock, _DUMMY_TEMP),
        call._test_node_._remove_temp_dir(_DUMMY_TEMP),
    ]


###############################################################################
###############################################################################
# CommandNode: _setup()

_SETUP_FILES_EXIST = (
    {"executables": ("ls", "sh")},
    {"input_files": _IN_FILES},
    {"auxiliary_files": _IN_FILES},
)


@pytest.mark.parametrize("kwargs", _INPUT_FILES_EXIST)
def test_commandnode_setup__files_exist(kwargs):
    cmd_mock = _build_cmd_mock(**kwargs)
    node = CommandNode(cmd_mock)
    node._setup(None, None)


_SETUP_FILES_MISSING = (
    {"executables": ("ls", "shxxxxxxxxxxx")},
    {"input_files": _OUT_FILES},
    {"auxiliary_files": _OUT_FILES},
)


@pytest.mark.parametrize("kwargs", _INPUT_FILES_MISSING)
def test_commandnode_setup__files_missing(kwargs):
    cmd_mock = _build_cmd_mock(**kwargs)
    node = CommandNode(cmd_mock)
    with pytest.raises(NodeError):
        node._setup(None, None)


###############################################################################
###############################################################################
# CommandNode: _run()


def test_commandnode_run__call_order():
    cmd_mock = _build_cmd_mock()
    node = CommandNode(cmd_mock)
    node._run(None, "xTMPx")

    assert cmd_mock.mock_calls == [call.run("xTMPx"), call.join()]


def test_commandnode_run__exception_on_error():
    cmd_mock = _build_cmd_mock(return_codes=(1,))
    node = CommandNode(cmd_mock)
    with pytest.raises(CmdNodeError):
        node._run(None, "xTMPx")

    assert cmd_mock.mock_calls == [call.run("xTMPx"), call.join()]


###############################################################################
###############################################################################
# CommandNode: _teardown


def _setup_temp_folders(tmp_path):
    destination = tmp_path / "dst"
    tmp_path = tmp_path / "tmp"
    tmp_path.mkdir(parents=True, exist_ok=True)
    destination.mkdir(parents=True, exist_ok=True)
    return destination, tmp_path


# Commit is called on the command obj
def test_commandnode_teardown__commit(tmp_path):
    cmd_mock = _build_cmd_mock()
    node = CommandNode(cmd_mock)
    node._teardown(None, tmp_path)
    assert cmd_mock.mock_calls == [call.commit(tmp_path)]


# Files exist in temp folder, and in destination after commit
def test_commandnode_teardown(tmp_path):
    destination, tmp_path = _setup_temp_folders(tmp_path)

    cmd = AtomicCmd(
        ("echo", "-n", "1 2 3"),
        IN_DUMMY=_EMPTY_FILE,
        OUT_STDOUT=str(destination / "foo.txt"),
    )
    cmd.run(tmp_path)
    assert cmd.join() == [0]
    node = CommandNode(cmd)
    assert (tmp_path / "foo.txt").exists()
    assert not (destination / "foo.txt").exists()
    node._teardown(None, tmp_path)
    assert not (tmp_path / "foo.txt").exists()
    assert (destination / "foo.txt").exists()


# Not all required files have been generated (atomic)
def test_commandnode_teardown__missing_files_in_temp(tmp_path):
    destination, tmp_path = _setup_temp_folders(tmp_path)

    cmd = AtomicCmd(
        ("echo", "-n", "1 2 3"),
        IN_DUMMY=_EMPTY_FILE,
        OUT_BAR=str(destination / "bar.txt"),
        OUT_STDOUT=str(destination / "foo.txt"),
    )
    cmd.run(tmp_path)
    assert cmd.join() == [0]
    node = CommandNode(cmd)
    temp_files_before = set(tmp_path.iterdir())
    dest_files_before = set(destination.iterdir())

    with pytest.raises(CmdNodeError):
        node._teardown(None, tmp_path)
    assert temp_files_before == set(tmp_path.iterdir())
    assert dest_files_before == set(destination.iterdir())


# Not all specified TEMP_ files exist at _teardown (allowed)
def test_commandnode_teardown__missing_optional_files(tmp_path):
    destination, tmp_path = _setup_temp_folders(tmp_path)

    cmd = AtomicCmd(
        ("echo", "-n", "1 2 3"),
        IN_DUMMY=_EMPTY_FILE,
        TEMP_OUT_BAR="bar.txt",
        OUT_STDOUT=str(destination / "foo.txt"),
    )
    cmd.run(tmp_path)
    assert cmd.join() == [0]
    node = CommandNode(cmd)
    node._teardown(None, tmp_path)
    assert os.listdir(str(tmp_path)) == []
    assert os.listdir(str(destination)) == ["foo.txt"]


# Not all required files were in place after commit
def test_commandnode_teardown__missing_files_in_dest(tmp_path):
    destination, tmp_path = _setup_temp_folders(tmp_path)

    class _CmdMock(AtomicCmd):
        def commit(self, temp):
            AtomicCmd.commit(self, temp)
            (destination / "foo.txt").unlink()

    cmd = _CmdMock(
        ("touch", "%(OUT_FOO)s", "%(OUT_BAR)s"),
        IN_DUMMY=_EMPTY_FILE,
        OUT_FOO=str(destination / "foo.txt"),
        OUT_BAR=str(destination / "bar.txt"),
    )
    cmd.run(tmp_path)
    assert cmd.join() == [0]
    node = CommandNode(cmd)
    with pytest.raises(NodeError):
        node._teardown(None, tmp_path)


def test_commandnode_teardown__extra_files_in_temp(tmp_path):
    destination, tmp_path = _setup_temp_folders(tmp_path)

    unexpected_file = tmp_path / "unexpected_file.txt"
    unexpected_file.write_text("1 2 3")

    cmd = AtomicCmd(
        ("echo", "1 2 3"),
        IN_DUMMY=_EMPTY_FILE,
        OUT_STDOUT=str(destination / "foo.txt"),
    )

    node = CommandNode(cmd)
    node._run(None, tmp_path)
    node._teardown(None, tmp_path)

    assert list(tmp_path.iterdir()) == [unexpected_file]
    assert list(destination.iterdir()) == [destination / "foo.txt"]



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import bz2
import errno
import gzip
import os
import shutil
import uuid
from pathlib import Path, PosixPath
from typing import IO, Any, Callable, Iterable, List, Optional, Tuple, Union

from .utilities import safe_coerce_to_tuple

try:
    import pathlib2

    PathClasses = (Path, PosixPath, pathlib2.PosixPath)
except ImportError:
    PathClasses = (Path, PosixPath)


try:
    fspath = os.fspath
except AttributeError:
    # Backwards compatibility with Python3.5
    def fspath(value):
        if isinstance(value, (str, bytes)):
            return value
        elif isinstance(value, PathClasses):
            return str(value)

        raise TypeError(
            "expected str/bytes/Path/PosixPath, not {}".format(type(value).__name__)
        )


def add_postfix(filename: Union[str, Path], postfix: str) -> str:
    """Ads a postfix to a filename (before any extensions that filename may have)."""
    filename, ext = os.path.splitext(filename)
    return filename + postfix + ext


def swap_ext(filename: str, ext: str) -> str:
    """Replaces the existing extension of a filename with the specified extension,
    ensuring that the extension is prefixed by a '.'. If no extension is specified,
    other than potentially a dot, the existing extension is stripped."""
    filename, _ = os.path.splitext(filename)
    if ext in ("", "."):
        return filename

    if not ext.startswith("."):
        ext = "." + ext

    return filename + ext


def reroot_path(root: Union[str, Path], filename: str) -> str:
    """Returns the basename of filename, joined to root."""
    return os.path.join(root, os.path.basename(filename))


def create_temp_dir(root: Union[str, Path]) -> str:
    """Creates a temporary directory, accessible only by the owner,
    at the specified location. The folder name is randomly generated,
    and only the current user has access"""

    def _generate_path() -> str:
        return os.path.join(fspath(root), str(uuid.uuid4()))

    path = _generate_path()
    while not make_dirs(path, mode=0o750):
        path = _generate_path()
    return path


def missing_files(filenames: Iterable[Union[str, Path]]) -> List[Union[str, Path]]:
    """Given a list of filenames, returns a list of those that
    does not exist. Note that this function does not differentiate
    between files and folders."""
    missing = []
    for filename in safe_coerce_to_tuple(filenames):
        filename = fspath(filename)
        if not os.path.exists(filename):
            missing.append(filename)
    return missing


def missing_executables(
    filenames: Iterable[Union[str, Path]],
) -> List[Union[str, Path]]:
    missing = []
    for filename in safe_coerce_to_tuple(filenames):
        filename = fspath(filename)
        if not shutil.which(filename):
            missing.append(filename)
    return missing


def make_dirs(directory: Union[str, Path], mode: int = 0o777) -> bool:
    """Wrapper around os.makedirs to make it suitable for using
    in a multithreaded/multiprocessing environment: Unlike the
    regular function, this wrapper does not throw an exception if
    the directory already exists, which may happen if another
    thread/process created the directory during the function call.

    Returns true if a new directory was created, false if it
    already existed. Other errors result in exceptions."""
    if not directory:
        raise ValueError("Empty directory passed to make_dirs()")

    try:
        os.makedirs(fspath(directory), mode=mode)
        return True
    except OSError as error:
        # make_dirs be called by multiple subprocesses at the same time,
        # so only raise if the actual creation of the folder failed
        if error.errno != errno.EEXIST:
            raise
        return False


def move_file(source: Union[str, Path], destination: Union[str, Path]) -> None:
    """Wrapper around shutils which ensures that the
    destination directory exists before moving the file."""
    _sh_wrapper(shutil.move, source, destination)


def copy_file(source: Union[str, Path], destination: Union[str, Path]) -> None:
    """Wrapper around shutils which ensures that the
    destination directory exists before copying the file."""
    _sh_wrapper(shutil.copy, source, destination)


def open_ro(filename: Union[str, Path], mode: str = "rt") -> IO[str]:
    """Opens a file for reading, transparently handling
    GZip and BZip2 compressed files. Returns a file handle."""
    filename = fspath(filename)
    if mode not in ("rt", "rb", "r"):
        raise ValueError(mode)
    elif mode == "r":
        # Ensure uniform behavior between open/gzip.open/bz2.open
        mode = "rt"

    with open(filename, "rb") as handle:
        header = handle.read(2)

    if header == b"\x1f\x8b":
        return gzip.open(filename, mode)
    elif header == b"BZ":
        return bz2.open(filename, mode)
    else:
        return open(filename, mode)


def try_remove(filename: Union[str, Path]) -> bool:
    """Tries to remove a file. Unlike os.remove, the function does not
    raise an exception if the file does not exist, but does raise
    exceptions on other errors. The return value reflects whether or
    not the file was actually removed."""
    return _try_rm_wrapper(os.remove, filename)


def try_rmtree(filename: Union[str, Path]) -> bool:
    """Tries to remove a dir-tree. Unlike shutil.rmtree, the function does not raise
    an exception if the file does not exist, but does raise exceptions on other
    errors. The return value reflects whether or not the file was actually
    removed."""
    return _try_rm_wrapper(shutil.rmtree, filename)


def describe_files(files: Iterable[str]) -> str:
    """Return a text description of a set of files."""
    files = validate_filenames(files)

    if not files:
        return "No files"
    elif len(files) == 1:
        return repr(files[0])

    glob_files = get_files_glob(files, max_differences=2)
    if glob_files:
        return repr(glob_files)

    paths = set(os.path.dirname(filename) for filename in files)
    if len(paths) == 1:
        return "%i files in '%s'" % (len(files), paths.pop())
    return "%i files" % (len(files),)


def describe_paired_files(files_1: Iterable[str], files_2: Iterable[str]) -> str:
    """Return a text description of a set of paired filenames; the
    sets must be of the same length, a the description will depend
    on the overall similarity of the filenames / paths. If 'files_2'
    is empty, this function is the equivalent of calling
    'describe_files' with 'files_1' as the argument. In all other
    cases the length of the two sets must be the same."""
    files_1 = validate_filenames(files_1)
    files_2 = validate_filenames(files_2)

    if files_1 and not files_2:
        return describe_files(files_1)
    elif len(files_1) != len(files_2):
        raise ValueError(
            "Unequal number of files for mate 1 vs mate 2 reads: %i vs %i"
            % (len(files_1), len(files_2))
        )

    glob_files_1 = get_files_glob(files_1, 3)
    glob_files_2 = get_files_glob(files_2, 3)
    if glob_files_1 and glob_files_2:
        final_glob = get_files_glob(
            (glob_files_1, glob_files_2), 1, show_differences=True
        )
        if final_glob:
            return repr(final_glob)

    fnames = files_1 + files_2
    paths = set(os.path.dirname(fname) for fname in fnames)
    if len(paths) == 1:
        return "%i pair(s) of files in '%s'" % (len(files_1), paths.pop())
    return "%i pair(s) of files" % (len(files_1),)


def get_files_glob(
    filenames: Iterable[str], max_differences: int = 1, show_differences: bool = False
) -> Optional[str]:
    """Tries to generate a glob-string for a set of filenames, containing
    at most 'max_differences' different columns. If more differences are
    found, or if the length of filenames vary, None is returned."""
    # File lengths must be the same, otherwise we'd have to do MSA
    if len(set(map(len, filenames))) > 1:
        return None

    glob_fname, differences = [], 0
    for chars in zip(*filenames):
        if "?" in chars:
            chars = ("?",)

        if len(frozenset(chars)) > 1:
            if show_differences:
                chars = ("[%s]" % ("".join(sorted(chars))),)
            else:
                chars = ("?",)
            differences += 1
        glob_fname.append(chars[0])

    if differences > max_differences:
        return None

    return "".join(glob_fname)


def validate_filenames(filenames: Iterable[str]) -> Tuple[str, ...]:
    """Sanity checks for filenames handled by
    'describe_files' and 'describe_paired_files."""
    return tuple(fspath(filename) for filename in safe_coerce_to_tuple(filenames))


def _sh_wrapper(
    func: Callable[[Union[str, Path], Union[str, Path]], Any],
    source: Union[str, Path],
    destination: Union[str, Path],
) -> None:
    """Runs an 'shutil' function ('func') which takes an 'source' and
    a 'destination' argument (e.g. copy/move/etc.), but silently
    handles the case where the destination directory does not exist.

    If this is the case, the function will first create the destination
    directory, and then retry the function."""
    source = fspath(source)
    destination = fspath(destination)

    try:
        func(source, destination)
    except IOError as error:
        if error.errno == errno.ENOENT:
            if source and destination and os.path.exists(source):
                dirname = os.path.dirname(destination)
                make_dirs(dirname)
                func(source, destination)
                return
        elif error.errno == errno.ENOSPC:
            # Not enough space; remove partial file
            os.unlink(destination)
        raise


def _try_rm_wrapper(func: Callable[[Any], Any], fpath: Union[str, Path]) -> bool:
    """Takes a function (e.g. os.remove / os.rmdir), and attempts to remove a
    path; returns true if that path was successfully remove, and false if it did
    not exist."""
    try:
        func(fspath(fpath))
        return True
    except OSError as error:
        if error.errno != errno.ENOENT:
            raise
        return False



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

import paleomix.common.fileutils as fileutils
from paleomix.nodes.mafft import MAFFTNode
from paleomix.nodes.sequences import CollectSequencesNode, FilterSingletonsNode


def build_msa_nodes(options, settings, regions, filtering, dependencies):
    if settings["Program"].lower() != "mafft":
        raise RuntimeError("Only MAFFT support has been implemented!")

    sequencedir = os.path.join(options.destination, "alignments", regions["Name"])
    # Run on full set of sequences
    sequences = regions["Sequences"][None]

    node = CollectSequencesNode(
        fasta_files=regions["Genotypes"],
        destination=sequencedir,
        sequences=sequences,
        dependencies=dependencies,
    )

    if settings["Enabled"]:
        fasta_files = {}
        algorithm = settings["MAFFT"]["Algorithm"]
        for sequence in sequences:
            input_file = os.path.join(sequencedir, sequence + ".fasta")
            output_file = os.path.join(sequencedir, sequence + ".afa")

            fasta_files[output_file] = MAFFTNode(
                input_file=input_file,
                output_file=output_file,
                algorithm=algorithm,
                options=settings["MAFFT"],
                dependencies=node,
            )
    else:
        fasta_files = dict((filename, node) for filename in node.output_files)

    if not any(filtering.values()):
        return list(fasta_files.values())

    destination = sequencedir + ".filtered"
    filtering = dict(filtering)
    filtered_nodes = []

    for filename, node in fasta_files.items():
        output_filename = fileutils.reroot_path(destination, filename)
        filtered_node = FilterSingletonsNode(
            input_file=filename,
            output_file=output_filename,
            filter_by=filtering,
            dependencies=node,
        )

        filtered_nodes.append(filtered_node)

    return filtered_nodes


def chain(_pipeline, options, makefiles):
    destination = options.destination  # Move to makefile
    for makefile in makefiles:
        nodes = []
        settings = makefile["MultipleSequenceAlignment"]
        filtering = makefile["Project"]["FilterSingletons"]
        options.destination = os.path.join(destination, makefile["Project"]["Title"])

        for regions in makefile["Project"]["Regions"].values():
            regions_settings = settings[regions["Name"]]
            nodes.extend(
                build_msa_nodes(
                    options, regions_settings, regions, filtering, makefile["Nodes"]
                )
            )

        makefile["Nodes"] = tuple(nodes)
    options.destination = destination



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import collections
import os
import random

from paleomix.common.fileutils import add_postfix, swap_ext
from paleomix.nodes.examl import ExaMLNode, ExaMLParserNode
from paleomix.nodes.formats import FastaToPartitionedInterleavedPhyNode as ToPhylipNode
from paleomix.nodes.newick import NewickRerootNode, NewickSupportNode
from paleomix.nodes.phylip import PHYLIPBootstrapNode
from paleomix.nodes.raxml import RAxMLParsimonyTreeNode


def _build_supermatrix(
    destination, input_files, exclude_samples, subset_files, dependencies
):
    matrixprefix = os.path.join(destination, "alignments")

    # This supermatrix (and partitions) is not used by the pipeline, but is
    # built for reference purposes; may be made optional in the future.
    supermatrix = ToPhylipNode(
        infiles=input_files,
        out_prefix=matrixprefix,
        exclude_groups=exclude_samples,
        dependencies=dependencies,
        file_dependencies=subset_files,
    )

    # Supermatrix with empty columns (all N, n, or -) stripped
    reduced = ToPhylipNode(
        reduce=True,
        infiles=input_files,
        out_prefix=matrixprefix + ".reduced",
        exclude_groups=exclude_samples,
        dependencies=dependencies,
        file_dependencies=subset_files,
    )

    return (supermatrix, reduced)


def _examl_nodes(
    options,
    settings,
    input_alignment,
    input_partitions,
    input_binary,
    output_template,
    dependencies,
):
    parsimony_tree = output_template % ("parsimony_tree",)
    tree = RAxMLParsimonyTreeNode(
        input_alignment=input_alignment,
        input_partitions=input_partitions,
        output_tree=parsimony_tree,
        dependencies=dependencies,
    )

    return ExaMLNode(
        input_binary=input_binary,
        initial_tree=parsimony_tree,
        output_template=output_template,
        model=settings["ExaML"]["Model"].upper(),
        threads=options.examl_max_threads,
        dependencies=tree,
    )


def _build_rerooted_trees(nodes, reroot_on):
    filenames = []
    for node in nodes:
        for filename in node.output_files:
            if filename.endswith(".result"):
                filenames.append(filename)

    output_file = os.path.dirname(filenames[0]) + ".newick"

    return NewickRerootNode(
        tree_files=filenames,
        output_file=output_file,
        taxa=reroot_on,
        dependencies=nodes,
    )


def _build_examl_replicates(
    options, phylo, destination, input_alignment, input_partition, dependencies
):
    input_binary = os.path.join(destination, "alignments.reduced.binary")
    binary = ExaMLParserNode(
        input_alignment=input_alignment,
        input_partition=input_partition,
        output_file=input_binary,
        dependencies=dependencies,
    )

    replicates = []
    for replicate_num in range(phylo["ExaML"]["Replicates"]):
        replicate_destination = os.path.join(destination, "replicates")
        replicate_template = os.path.join(
            replicate_destination, "replicate.%04i.%%s" % (replicate_num,)
        )
        replicates.append(
            _examl_nodes(
                options,
                phylo,
                input_alignment,
                input_partition,
                input_binary,
                replicate_template,
                binary,
            )
        )

    if replicates:
        return _build_rerooted_trees(replicates, phylo["RootTreesOn"])

    return None


def _build_examl_bootstraps(
    options, phylo, destination, input_alignment, input_partition, dependencies
):
    bootstraps = []
    num_bootstraps = phylo["ExaML"]["Bootstraps"]
    bootstrap_destination = os.path.join(destination, "bootstraps")
    bootstrap_template = os.path.join(bootstrap_destination, "bootstrap.%04i.phy")

    for bootstrap_num in range(num_bootstraps):
        bootstrap_alignment = bootstrap_template % (bootstrap_num,)
        bootstrap = PHYLIPBootstrapNode(
            input_alignment=input_alignment,
            input_partition=input_partition,
            output_alignment=bootstrap_alignment,
            seed=random.randint(1, 2**32 - 1),
            dependencies=dependencies,
        )

        bootstrap_binary = swap_ext(bootstrap_alignment, ".binary")
        bootstrap_final = swap_ext(bootstrap_alignment, ".%s")
        bs_binary = ExaMLParserNode(
            input_alignment=bootstrap_alignment,
            input_partition=input_partition,
            output_file=bootstrap_binary,
            dependencies=bootstrap,
        )

        bootstraps.append(
            _examl_nodes(
                options=options,
                settings=phylo,
                input_alignment=bootstrap_alignment,
                input_partitions=input_partition,
                input_binary=bootstrap_binary,
                output_template=bootstrap_final,
                dependencies=bs_binary,
            )
        )

    if bootstraps:
        return _build_rerooted_trees(bootstraps, phylo["RootTreesOn"])

    return None


def add_bootstrap_support(destination, replicate, bootstrap):
    if not (replicate and bootstrap):
        return [_f for _f in (replicate, bootstrap) if _f]

    replicate_file = os.path.join(destination, "replicates.newick")
    bootstrap_file = os.path.join(destination, "bootstraps.newick")
    output_file = add_postfix(replicate_file, ".support")

    return (
        NewickSupportNode(
            main_tree_files=replicate_file,
            support_tree_files=bootstrap_file,
            output_file=output_file,
            dependencies=(bootstrap, replicate),
        ),
    )


def _build_examl_nodes(
    options, settings, destination, input_files, subset_files, dependencies
):
    input_alignment = os.path.join(destination, "alignments.reduced.phy")
    input_partition = os.path.join(destination, "alignments.reduced.partitions")

    excluded = settings["ExcludeSamples"]
    supermatrix = _build_supermatrix(
        destination, input_files, excluded, subset_files, dependencies
    )
    examl_args = (
        options,
        settings,
        destination,
        input_alignment,
        input_partition,
        supermatrix,
    )

    examl_replicates = _build_examl_replicates(*examl_args)
    examl_bootstraps = _build_examl_bootstraps(*examl_args)
    examl_dependencies = add_bootstrap_support(
        destination, examl_replicates, examl_bootstraps
    )

    return examl_dependencies


def _build_examl_per_gene_nodes(
    options, settings, run_dd, roi, destination, filtering, dependencies
):
    regions = settings["Project"]["Regions"][roi["Name"]]
    sequences = regions["Sequences"][roi["SubsetRegions"]]
    subset_files = regions["SubsetFiles"][roi["SubsetRegions"]]
    filtering_postfix = ".filtered" if any(filtering.values()) else ""
    sequence_dir = os.path.join(
        options.destination, "alignments", roi["Name"] + filtering_postfix
    )
    msa_enabled = settings["MultipleSequenceAlignment"][regions["Name"]]["Enabled"]
    fasta_extension = ".afa" if msa_enabled else ".fasta"

    partitions = roi["Partitions"] or "111"

    nodes = []
    for sequence in sequences:
        seq_source = os.path.join(sequence_dir, sequence + fasta_extension)
        seq_destination = os.path.join(destination, sequence)
        input_files = {sequence: {"partitions": partitions, "filenames": [seq_source]}}
        nodes.extend(
            _build_examl_nodes(
                options,
                run_dd,
                seq_destination,
                input_files,
                subset_files,
                dependencies,
            )
        )

    return nodes


def _build_examl_regions_nodes(
    options, settings, run_dd, destination, filtering, dependencies
):
    input_files = collections.defaultdict(dict)
    subset_files = []
    for roi_name, roi_dd in run_dd["RegionsOfInterest"].items():
        regions = settings["Project"]["Regions"][roi_name]
        subset_key = roi_dd.get("SubsetRegions")
        sequences = regions["Sequences"][subset_key]
        subset_files.extend(regions["SubsetFiles"][subset_key])

        partitions = roi_dd["Partitions"]
        filtering_postfix = ".filtered" if any(filtering.values()) else ""
        sequence_dir = os.path.join(
            options.destination, "alignments", roi_name + filtering_postfix
        )
        msa_enabled = settings["MultipleSequenceAlignment"][regions["Name"]]["Enabled"]
        fasta_extension = ".afa" if msa_enabled else ".fasta"

        if partitions:
            for sequence in sequences:
                seq_source = os.path.join(sequence_dir, sequence + fasta_extension)

                input_files[sequence] = {
                    "partitions": partitions,
                    "filenames": [seq_source],
                }
        else:
            filenames = []
            for sequence in sequences:
                filenames.append(os.path.join(sequence_dir, sequence + fasta_extension))

            input_files[roi_name] = {"partitions": "1", "filenames": filenames}

    return _build_examl_nodes(
        options, run_dd, destination, dict(input_files), subset_files, dependencies
    )


def build_phylogeny_nodes(options, settings, filtering, dependencies):
    nodes = []
    for run_name, run_dd in settings["PhylogeneticInference"].items():
        destination = os.path.join(options.destination, "phylogenies", run_name)

        if run_dd["PerGeneTrees"]:
            run_nodes = []
            for roi in run_dd["RegionsOfInterest"].values():
                roi_destination = os.path.join(destination, roi["Name"])
                run_nodes.extend(
                    _build_examl_per_gene_nodes(
                        options,
                        settings,
                        run_dd,
                        roi,
                        roi_destination,
                        filtering,
                        dependencies,
                    )
                )
            nodes.extend(run_nodes)
        else:
            nodes.extend(
                _build_examl_regions_nodes(
                    options, settings, run_dd, destination, filtering, dependencies
                )
            )

    return nodes


def chain_examl(_pipeline, options, makefiles):
    destination = options.destination  # Move to makefile
    for makefile in makefiles:
        filtering = makefile["Project"]["FilterSingletons"]
        options.destination = os.path.join(destination, makefile["Project"]["Title"])

        makefile["Nodes"] = build_phylogeny_nodes(
            options, makefile, filtering, makefile["Nodes"]
        )
    options.destination = destination



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import collections
import os
import re
import signal
import sys
import weakref

import paleomix.atomiccmd.pprint as atomicpp
import paleomix.common.fileutils as fileutils
import paleomix.common.procs as procs
from paleomix.common.utilities import safe_coerce_to_tuple

_PIPES = (("IN", "IN_STDIN"), ("OUT", "OUT_STDOUT"), ("OUT", "OUT_STDERR"))
_KEY_RE = re.compile("^(IN|OUT|EXEC|AUX|CHECK|TEMP_IN|TEMP_OUT)_[A-Z0-9_]+")


class CmdError(RuntimeError):
    """Exception raised for AtomicCmd specific errors."""

    def __init__(self, msg):
        RuntimeError.__init__(self, msg)


class AtomicCmd:
    """Executes a command, only moving resulting files to the destination
    directory if the command was successful. This helps prevent the
    accidental use of partial files in downstream analysis, and eases
    restarting of a pipeline following errors (no cleanup).

    Inividual files/paths in the command are specified using keywords (see
    the documentation for the constructor), allowing the command to be
    transparently modified to execute in a temporary directory.

    When an AtomicCmd is run(), a signal handler is installed for SIGTERM,
    which ensures that any running processes are terminated. In the absence
    of this, AtomicCmds run in terminated subprocesses can result in still
    running children after the termination of the parents."""

    PIPE = procs.PIPE
    DEVNULL = procs.DEVNULL

    def __init__(self, command, set_cwd=False, **kwargs):
        """Takes a command and a set of files.

        The command is expected to be an iterable starting with the name of an
        executable, with each item representing one string on the command line.
        Thus, the command "find /etc -name 'profile*'" might be represented as
        the list ["find", "/etc", "-name", "profile*"].

        Commands typically consist of an executable, one or more input files,
        one or more output files, and one or more pipes. In atomic command,
        such files are not specified directly, but instead are specified using
        keywords, which allows easy tracking of requirements and other
        features. Note that only files, and not directories, are supported as
        input/output!

        Each keyword represents a type of file, as determined by the prefix:
           IN_    -- Path to input file transformed/analysed the executable.
           OUT_   -- Path to output file generated by the executable. During
                     execution of the AtomicCmd, these paths are modified to
                     point to the temporary directory.
           EXEC_  -- Name of / path to executable. The first item in the
                     command is always one of the executables, even if not
                     specified in this manner.
           AUX_   -- Auxiliary files required by the executable(s), which are
                     themselves not executable. Examples include scripts,
                     config files, data-bases, and the like.
           CHECK_ -- A callable, which upon calling does version checking,
                     raising an exception in the case of requirements not being
                     met. This may be used to ensure that prerequisites are met
                     before running the command. The function is not called by
                     AtomicCmd itself.

        EXAMPLE 1: Creating a gzipped tar-archive from two files
        The command "tar cjf output-file input-file-1 input-file-2" could be
        represented using the following AtomicCmd:
        cmd = AtomicCmd(["tar", "cjf", "%(OUT_FILE)s",
                         "%(IN_FILE_1)s", "%(IN_FILE_2)s"],
                        OUT_FILE  = "output-file",
                        IN_FILE_1 = "input-file-1",
                        IN_FILE_2 = "input-file-2")

        Note that files that are not directly invoked may be included above,
        in order to allow the specification of requirements. This could include
        required data files, or executables indirectly executed by a script.

        If the above is prefixed with "TEMP_", files are read from / written
        to the temporary folder in which the command is executed. Note that all
        TEMP_OUT_ files are deleted when commit is called (if they exist), and
        only filenames (not dirname component) are allowed for TEMP_ values.

        In addition, the follow special names may be used with the above:
           STDIN_  -- Takes a filename, or an AtomicCmd, in which case stdout
                      of that command is piped to the stdin of this instance.
           STDOUT_ -- Takes a filename, or the special value PIPE to allow
                      another AtomicCmd instance to use the output directly.
           STDERR_ -- Takes a filename.

        Each pipe can only be used once, with or without the TEMP_ prefix.

        EXAMPLE 2: zcat'ing an archive
        The command "zcat input-file > output-file" could be represented using
        the following AtomicCmd:
        cmd = AtomicCmd(["zcat", "%(IN_FILE)s"],
                        OUT_STDOUT = "output-file")

        If 'set_cwd' is True, the current working directory is set to the
        temporary directory before the command is executed. Input paths are
        automatically turned into absolute paths in this case."""
        self._proc = None
        self._temp = None
        self._running = False
        self._command = list(map(str, safe_coerce_to_tuple(command)))
        self._set_cwd = set_cwd
        self._terminated = False

        if not self._command or not self._command[0]:
            raise ValueError("Empty command in AtomicCmd constructor")

        arguments = self._process_arguments(id(self), self._command, kwargs)
        self._files = self._build_files_dict(arguments)

        file_sets = self._build_files_map(self._command, arguments)
        self.executables = file_sets["EXEC"]
        self.requirements = file_sets["CHECK"]
        self.input_files = file_sets["IN"]
        self.output_files = file_sets["OUT"]
        self.auxiliary_files = file_sets["AUX"]
        self.optional_temp_files = file_sets["TEMP_OUT"]
        self.expected_temp_files = frozenset(
            os.path.basename(path) for path in file_sets["OUT"]
        )

        # Dry-run, to catch errors early
        self._generate_call("/tmp")

    def run(self, temp, wrap_errors=True):
        """Runs the given command, saving files in the specified temp folder.
        To move files to their final destination, call commit(). Note that in
        contexts where the *Cmds classes are used, this function may block.

        """
        if self._running:
            raise CmdError("Calling 'run' on already running command.")

        temp = fileutils.fspath(temp)
        self._temp = temp
        self._running = True

        # kwords for pipes are always built relative to the current directory,
        # since these are opened before (possibly) CD'ing to the temp
        # directory.
        stdin = stdout = stderr = None
        try:
            kwords = self._generate_filenames(self._files, root=temp)
            stdin = self._open_pipe(kwords, "IN_STDIN", "rb")
            stdout = self._open_pipe(kwords, "OUT_STDOUT", "wb")
            stderr = self._open_pipe(kwords, "OUT_STDERR", "wb")

            cwd = temp if self._set_cwd else None
            temp = "" if self._set_cwd else os.path.abspath(temp)
            call = self._generate_call(temp)

            # Explicitly set to DEVNULL to ensure that STDIN is not left open.
            if stdin is None:
                stdin = self.DEVNULL

            self._proc = procs.open_proc(
                call,
                stdin=stdin,
                stdout=stdout,
                stderr=stderr,
                cwd=cwd,
                preexec_fn=os.setsid,
            )
        except Exception as error:
            if not wrap_errors:
                raise

            message = "Error running commands:\n  Call = %r\n  Error = %r"
            raise CmdError(message % (self._command, error))
        finally:
            # Close pipes to allow the command to receive SIGPIPE
            for handle in (stdin, stdout, stderr):
                if handle not in (None, self.PIPE, self.DEVNULL):
                    handle.close()

        # Allow subprocesses to be killed in case of a SIGTERM
        _add_to_killlist(self._proc)

    def ready(self):
        """Returns true if the command has been run to completion,
        regardless of whether or not an error occurred."""
        return self._proc and self._proc.poll() is not None

    def join(self):
        """Similar to Popen.wait(), but returns the value wrapped in a list,
        and ensures that any opened handles are closed. Must be called before
        calling commit."""
        if not self._proc:
            return [None]

        self._running = False
        return_code = self._proc.wait()
        if return_code < 0:
            return_code = signal.Signals(-return_code).name
        return [return_code]

    def wait(self):
        """Equivalent to Subproces.wait. This function should only
        be used in contexts where a AtomicCmd needs to be combined
        with Subprocesses, as it does not exist for AtomicSets."""
        return self.join()[0]

    def terminate(self):
        """Sends SIGTERM to process if it is still running.
        Has no effect if the command has already finished."""
        if self._proc and self._proc.poll() is None:
            try:
                os.killpg(self._proc.pid, signal.SIGTERM)
                self._terminated = True
            except OSError:
                pass  # Already dead / finished process

    def commit(self, temp):
        temp = fileutils.fspath(temp)
        if not self.ready():
            raise CmdError("Attempting to commit before command has completed")
        elif self._running:
            raise CmdError("Called 'commit' before calling 'join'")
        elif not os.path.samefile(self._temp, temp):
            raise CmdError(
                "Mismatch between previous and current temp folders"
                ": %r != %s" % (self._temp, temp)
            )

        missing_files = self.expected_temp_files - set(os.listdir(temp))
        if missing_files:
            raise CmdError(
                "Expected files not created: %s" % (", ".join(missing_files))
            )

        temp = os.path.abspath(temp)
        filenames = self._generate_filenames(self._files, temp)
        committed_files = set()
        try:
            for key, filename in filenames.items():
                if isinstance(filename, str):
                    if key.startswith("OUT_"):
                        fileutils.move_file(filename, self._files[key])
                        committed_files.add(self._files[key])
                    elif key.startswith("TEMP_OUT_"):
                        fileutils.try_remove(filename)
        except Exception:
            # Cleanup after failed commit
            for fpath in committed_files:
                fileutils.try_remove(fpath)
            raise

        self._proc = None
        self._temp = None

    def __str__(self):
        return atomicpp.pformat(self)

    def _generate_call(self, temp):
        kwords = self._generate_filenames(self._files, root=temp)

        try:
            return [(field % kwords) for field in self._command]
        except (TypeError, ValueError) as error:
            raise CmdError(
                "Error building Atomic Command:\n"
                "  Call = %s\n  Error = %s: %s"
                % (self._command, error.__class__.__name__, error)
            )
        except KeyError as error:
            raise CmdError(
                "Error building Atomic Command:\n"
                "  Call = %s\n  Value not specified for path = %s"
                % (self._command, error)
            )

    @classmethod
    def _process_arguments(cls, proc_id, command, kwargs):
        arguments = collections.defaultdict(dict)
        for key, value in kwargs.items():
            match = _KEY_RE.match(key)
            if not match:
                raise ValueError("Invalid keyword argument %r" % (key,))

            # None is ignored, to make use of default arguments easier
            if value is not None:
                (group,) = match.groups()
                arguments[group][key] = value

        # Pipe stdout/err to files by default
        executable = os.path.basename(command[0])
        for pipe in ("STDOUT", "STDERR"):
            has_out_pipe = ("OUT_" + pipe) in arguments["OUT"]
            has_temp_out_pipe = ("TEMP_OUT_" + pipe) in arguments["TEMP_OUT"]
            if not (has_out_pipe or has_temp_out_pipe):
                filename = "pipe_%s_%i.%s" % (executable, proc_id, pipe.lower())
                arguments["TEMP_OUT"]["TEMP_OUT_" + pipe] = filename

        cls._validate_arguments(arguments)
        cls._validate_output_files(arguments)
        cls._validate_pipes(arguments)

        return arguments

    @classmethod
    def _validate_arguments(cls, arguments):
        # Output files
        for group in ("OUT", "TEMP_OUT"):
            for key, value in arguments.get(group, {}).items():
                if isinstance(value, str):
                    continue

                if key in ("OUT_STDOUT", "TEMP_OUT_STDOUT"):
                    if value not in (cls.PIPE, cls.DEVNULL):
                        raise TypeError(
                            "STDOUT must be a string, PIPE "
                            "or DEVNULL, not %r" % (value,)
                        )
                elif key in ("OUT_STDERR", "TEMP_OUT_STDERR"):
                    if value is not cls.DEVNULL:
                        raise TypeError(
                            "STDERR must be a string, or DEVNULL, not %r" % (value,)
                        )
                else:
                    raise TypeError("%s must be string, not %r" % (key, value))

        # Input files, including executables and auxiliary files
        for group in ("IN", "TEMP_IN", "EXEC", "AUX"):
            for key, value in arguments.get(group, {}).items():
                if isinstance(value, str):
                    continue

                if key in ("IN_STDIN", "TEMP_IN_STDIN"):
                    if not isinstance(value, AtomicCmd) and value is not cls.DEVNULL:
                        raise TypeError(
                            "STDIN must be string, AtomicCmd, "
                            "or DEVNULL, not %r" % (value,)
                        )
                else:
                    raise TypeError("%s must be string, not %r" % (key, value))

        for key, value in arguments.get("CHECK", {}).items():
            if not callable(value):
                raise TypeError("%s must be callable, not %r" % (key, value))

        for group in ("TEMP_IN", "TEMP_OUT"):
            for key, value in arguments.get(group, {}).items():
                is_string = isinstance(value, str)
                if is_string and os.path.dirname(value):
                    raise ValueError(
                        "%s cannot contain dir component: %r" % (key, value)
                    )

        return True

    @classmethod
    def _validate_output_files(cls, arguments):
        output_files = collections.defaultdict(list)
        for group in ("OUT", "TEMP_OUT"):
            for key, value in arguments.get(group, {}).items():
                if isinstance(value, str):
                    filename = os.path.basename(value)
                    output_files[filename].append(key)

        for filename, keys in output_files.items():
            if len(keys) > 1:
                raise ValueError(
                    "Same output filename (%s) is specified for "
                    "multiple keys: %s" % (filename, ", ".join(sorted(keys)))
                )

    @classmethod
    def _validate_pipes(cls, arguments):
        for group, pipe in _PIPES:
            has_pipe = pipe in arguments[group]
            has_temp_pipe = ("TEMP_" + pipe) in arguments["TEMP_" + group]
            if has_pipe and has_temp_pipe:
                raise CmdError("Pipes may only be specified once")

    @classmethod
    def _open_pipe(cls, kwords, pipe, mode):
        filename = kwords.get(pipe, kwords.get("TEMP_" + pipe))
        if filename in (None, cls.PIPE, cls.DEVNULL):
            return filename
        elif isinstance(filename, AtomicCmd):
            return filename._proc and filename._proc.stdout

        return open(filename, mode)

    @classmethod
    def _generate_filenames(cls, files, root):
        filenames = {"TEMP_DIR": root}
        for key, filename in files.items():
            if isinstance(filename, str):
                if key.startswith("TEMP_") or key.startswith("OUT_"):
                    filename = os.path.join(root, os.path.basename(filename))
                elif not root and (key.startswith("IN_") or key.startswith("AUX_")):
                    filename = os.path.abspath(filename)
            filenames[key] = filename

        return filenames

    @classmethod
    def _build_files_dict(cls, arguments):
        files = {}
        for groups in arguments.values():
            for key, value in groups.items():
                files[key] = value

        return files

    @classmethod
    def _build_files_map(cls, command, arguments):
        file_sets = {
            "AUX": set(),
            "CHECK": set(),
            "EXEC": set(command[:1]),
            "IN": set(),
            "OUT": set(),
            "TEMP_IN": set(),
            "TEMP_OUT": set(),
        }

        for group, files in arguments.items():
            for key, filename in files.items():
                if isinstance(filename, str) or group == "CHECK":
                    file_sets[group].add(filename)

        return {key: frozenset(value) for key, value in file_sets.items()}


# The following ensures proper cleanup of child processes, for example in the
# case where multiprocessing.Pool.terminate() is called.
_PROCS = None


def _cleanup_children(signum, _frame):
    for proc_ref in list(_PROCS):
        proc = proc_ref()
        if proc:
            try:
                os.killpg(proc.pid, signal.SIGTERM)
            except OSError:
                # Ignore already closed processes, etc.
                pass
    sys.exit(-signum)


def _add_to_killlist(proc):
    global _PROCS

    if _PROCS is None:
        signal.signal(signal.SIGTERM, _cleanup_children)
        _PROCS = set()

    _PROCS.add(weakref.ref(proc, _PROCS.remove))



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import bisect
import random


def weighted_sampling(choices, weights, rng=random):
    choices = list(choices)
    weights = list(weights)
    if not weights or (len(weights) != len(choices)):
        raise ValueError(
            "Choices and probabilities must be non-empty lists "
            "of identical length, not lengths %i and %i" % (len(choices), len(weights))
        )

    total = 0
    totals = []
    for index, weight in enumerate(weights, start=1):
        if weight <= 0:
            raise ValueError(
                "Probabilities must be > 0, not %r for weight %i" % (weight, index)
            )
        total += weight
        totals.append(total)

    while True:
        rand = rng.random() * total
        index = bisect.bisect_right(totals, rand)
        yield choices[index]


def reservoir_sampling(items, downsample_to, rng=random):
    if not isinstance(downsample_to, int):
        raise TypeError(
            "Unexpected type for 'downsample_to': %r" % (type(downsample_to),)
        )
    elif downsample_to < 0:
        raise ValueError("Negative value for 'downsample_to': %i" % (downsample_to,))

    reservoir = []
    for index, item in enumerate(items):
        if index >= downsample_to:
            index = rng.randint(0, index)
            if index < downsample_to:
                reservoir[index] = item
        else:
            reservoir.append(item)
    return reservoir



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import json
import sys

from paleomix.common.argparse import ArgumentParser
from paleomix.common.formats.fastq import FASTQ, FASTQualities


def parse_args(argv):
    parser = ArgumentParser("paleomix :validate_fastq")
    parser.add_argument("files", nargs="+")
    parser.add_argument("--collapsed", action="store_true")
    parser.add_argument("--no-empty", action="store_true")
    parser.add_argument("--offset", type=int, choices=(33, 64), default=33)

    return parser.parse_args(argv)


def main(argv):
    args = parse_args(argv)

    seq_retained_nts = 0
    seq_retained_reads = 0

    for filename in args.files:
        qualities = FASTQualities()
        for record in FASTQ.from_file(filename):
            qualities.update(record)

            seq_retained_reads += 1
            seq_retained_nts += len(record.sequence)

        offsets = qualities.offsets()
        if offsets == FASTQualities.BOTH:
            print(
                "FASTQ file(s) contains quality scores with both quality offsets (33 "
                "and 64); file may be unexpected format or corrupt. Please ensure that "
                "this file contains valid FASTQ reads from a single source.",
                file=sys.stderr,
            )

            return 1
        elif offsets == FASTQualities.MISSING:
            if args.no_empty:
                print("FASTQ file is empty.", file=sys.stderr)

                return 1
        elif offsets not in (FASTQualities.AMBIGUOUS, args.offset):
            print(
                "FASTQ file contains quality scores with wrong quality score offset "
                "(%i); expected reads with quality score offset %i. Ensure that the "
                "'QualityOffset' specified in the makefile corresponds to the input."
                % (offsets, args.offset),
                file=sys.stderr,
            )

            return 1

    print(
        json.dumps(
            {
                "filenames": args.files,
                "seq_retained_reads": seq_retained_reads,
                "seq_retained_nts": seq_retained_nts,
                "seq_collapsed": seq_retained_reads if args.collapsed else 0,
            },
            indent=2,
            sort_keys=True,
        )
    )

    return 0


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import copy
import sys

from paleomix.common.bamfiles import BAMRegionsIter
from paleomix.common.timer import BAMTimer
from paleomix.common.utilities import get_in, set_in
from paleomix.tools.bam_stats.common import (
    collect_readgroups,
    collect_references,
    main_wrapper,
)
from paleomix.tools.bam_stats.coverage import ReadGroup, write_table

##############################################################################
##############################################################################


def build_region_template(args, handle):
    template = {}
    for key in collect_readgroups(args, handle):
        template[key] = ReadGroup()
    return template


def get_region_table(counts, region, template):
    subtable = counts.get(region)
    if subtable is not None:
        return subtable

    subtable = copy.deepcopy(template)
    counts[region] = subtable

    return subtable


##############################################################################
##############################################################################


def create_or_get_subtable(table, subtable_key, size):
    subtable = get_in(table, subtable_key)
    if subtable is None:
        subtable = ReadGroup()
        subtable.Size = size
        set_in(table, subtable_key, subtable)
    return subtable


def build_table(args, handle, counts):
    references = collect_references(args, handle)

    table = {}
    for key, readgroup in collect_readgroups(args, handle).items():
        sample = readgroup["SM"]
        library = readgroup["LB"]

        # Exclude counts for reads with no read-groups, if none such were seen
        if key is None and not args.ignore_readgroups:
            for reference in references:
                subtable = counts.get(reference, {}).get(key)
                if subtable is not None and subtable.has_values():
                    break
            else:
                continue

        for reference, size in references.items():
            subtable_key = (args.target_name, sample, library, reference)
            subtable = create_or_get_subtable(table, subtable_key, size)

            statistics = counts.get(reference, {}).get(key)
            if statistics is not None:
                subtable.add(statistics)

    return table


def print_table(args, handle, counts):
    table = build_table(args, handle, counts)
    write_table(table, args.outfile)


##############################################################################
##############################################################################


def process_record(subtable, record, flags, region):
    qname = record.qname
    if qname.startswith("M_") or qname.startswith("MT_"):
        subtable.Collapsed += 1
    elif flags & 0x40:  # first of pair
        subtable.PE_1 += 1
    elif flags & 0x80:  # second of pair
        subtable.PE_2 += 1
    else:  # Singleton
        subtable.SE += 1

    position = record.pos
    start = region.start
    end = region.end

    for cigar, num in record.cigar:
        left = min(max(position, start), end)
        right = min(max(position + num, start), end)
        bases_in_region = right - left
        assert 0 <= bases_in_region <= num

        # 0 = 'M', 1 = 'I', 2 = 'D', 7 = '=', 8 = 'X'
        if cigar in (0, 1, 2, 7, 8):
            if bases_in_region:
                subtable["MID    MM"[cigar]] += bases_in_region

            if cigar != 1:  # Everything but insertions
                position += num
        elif cigar == 3:  # N
            position += num


def process_file(handle, args):
    timer = BAMTimer(handle, step=1000000)

    counts = {}
    last_tid = 0
    region_template = build_region_template(args, handle)
    for region in BAMRegionsIter(handle, args.regions):
        if region.name is None:
            # Trailing unmapped reads
            break

        name = region.name
        if not args.regions and (handle.nreferences > args.max_contigs):
            name = "<Genome>"

        last_pos = 0
        region_table = get_region_table(counts, name, region_template)
        for position, records in region:
            for record in records:
                readgroup = args.get_readgroup_func(record)
                readgroup_table = region_table.get(readgroup)
                if readgroup_table is None:
                    # Unknown readgroups are treated as missing readgroups
                    readgroup_table = region_table[None]

                process_record(readgroup_table, record, record.flag, region)
                timer.increment(read=record)

            if (region.tid, position) < (last_tid, last_pos):
                sys.stderr.write("ERROR: Input BAM file is unsorted\n")
                return 1

            last_pos = position
            last_tid = region.tid

    timer.finalize()

    print_table(args, handle, counts)

    return 0


def main(argv):
    return main_wrapper(process_file, argv, ".coverage")


##############################################################################
##############################################################################

if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is herby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os
from unittest.mock import Mock

from paleomix.common.fileutils import fspath
from paleomix.nodegraph import FileStatusCache, NodeGraph

_TIMESTAMP_1 = 1000190760
_TIMESTAMP_2 = 1120719000


def create_test_file(utime, *args):
    filename = os.path.join(*map(fspath, args))
    with open(filename, "wb"):
        pass

    os.utime(filename, (utime, utime))

    return filename


###############################################################################
###############################################################################
# NodeGraph: _is_done
# TODO: Avoid testing private function, mock cache


def test_nodegraph_is_done__no_output():
    cache = FileStatusCache()
    node = Mock(output_files=())
    assert NodeGraph.is_done(node, cache)


def test_nodegraph_is_done__output_changes(tmp_path):
    temp_file_1 = tmp_path / "file_1.txt"
    temp_file_2 = tmp_path / "file_2.txt"
    my_node = Mock(output_files=(str(temp_file_1), str(temp_file_2)))
    assert not NodeGraph.is_done(my_node, FileStatusCache())
    temp_file_1.write_text("foo")
    assert not NodeGraph.is_done(my_node, FileStatusCache())
    temp_file_2.write_text("bar")
    assert NodeGraph.is_done(my_node, FileStatusCache())


def test_nodegraph_is_done__subnode_not_considered(tmp_path):
    temp_file = tmp_path / "file.txt"
    subnode = Mock(output_files=(str(temp_file),))
    my_node = Mock(output_files=(), subnodes=(subnode,))
    assert NodeGraph.is_done(my_node, FileStatusCache())


def test_nodegraph_is_outdated__no_output():
    my_node = Mock(input_files=(), output_files=())
    assert not NodeGraph.is_outdated(my_node, FileStatusCache())


def test_nodegraph_is_outdated__input_but_no_output(tmp_path):
    input_file = tmp_path / "file"
    input_file.touch()

    my_node = Mock(input_files=(str(input_file),), output_files=())
    assert not NodeGraph.is_outdated(my_node, FileStatusCache())


def test_nodegraph_is_outdated__output_but_no_input(tmp_path):
    output_file = tmp_path / "file"
    output_file.touch()

    my_node = Mock(input_files=(), output_files=(str(output_file),))
    assert not NodeGraph.is_outdated(my_node, FileStatusCache())


def test_nodegraph_is_outdated__not_outdated(tmp_path):
    my_node = Mock(
        input_files=(create_test_file(_TIMESTAMP_1, tmp_path, "older_file"),),
        output_files=(create_test_file(_TIMESTAMP_2, tmp_path, "younger_file"),),
    )
    assert not NodeGraph.is_outdated(my_node, FileStatusCache())


def test_nodegraph_is_outdated__outdated(tmp_path):
    my_node = Mock(
        input_files=(create_test_file(_TIMESTAMP_2, tmp_path, "younger_file"),),
        output_files=(create_test_file(_TIMESTAMP_1, tmp_path, "older_file"),),
    )
    assert NodeGraph.is_outdated(my_node, FileStatusCache())


def test_nodegraph_is_outdated__updates(tmp_path):
    older_file = create_test_file(_TIMESTAMP_1, tmp_path, "older_file")
    younger_file = create_test_file(_TIMESTAMP_2, tmp_path, "younger_file")

    my_node = Mock(
        input_files=(older_file,),
        output_files=(younger_file,),
    )
    assert not NodeGraph.is_outdated(my_node, FileStatusCache())
    my_node = Mock(
        input_files=(younger_file,),
        output_files=(older_file,),
    )
    assert NodeGraph.is_outdated(my_node, FileStatusCache())



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

import paleomix.common.fileutils as fileutils
import paleomix.common.versions as versions
from paleomix.atomiccmd.builder import (
    AtomicCmdBuilder,
    apply_options,
)
from paleomix.node import CommandNode

_VERSION_2_CHECK = versions.Requirement(
    call=("AdapterRemoval", "--version"),
    search=r"ver. (\d+)\.(\d+)\.(\d+)",
    checks=versions.GE(2, 2, 0),
)

_VERSION_3_CHECK = versions.Requirement(
    call=("adapterremoval3", "--version"),
    search=r"v(\d+)\.(\d+)\.(\d+)",
    checks=versions.GE(3, 0, 0),
)


class SE_AdapterRemoval2Node(CommandNode):
    def __init__(
        self, input_file, output_prefix, threads=1, options={}, dependencies=()
    ):
        # See below for parameters in common between SE/PE
        cmd = AtomicCmdBuilder("AdapterRemoval", CHECK_VERSION=_VERSION_2_CHECK)
        cmd = _set_common_parameters(cmd, threads=threads, options=options)

        # Prefix for output files, ensure that all end up in temp folder
        cmd.set_option("--basename", "%(TEMP_OUT_BASENAME)s")

        output_tmpl = output_prefix + ".%s.gz"
        cmd.set_kwargs(
            TEMP_OUT_BASENAME=os.path.basename(output_prefix),
            OUT_SETTINGS=output_prefix + ".settings",
            OUT_MATE_1=output_tmpl % ("truncated",),
            OUT_DISCARDED=output_tmpl % ("discarded",),
        )

        cmd.set_option("--file1", "%(IN_READS_1)s")
        cmd.set_kwargs(IN_READS_1=input_file)

        apply_options(cmd, options)

        self.files = {
            "Statistics": output_prefix + ".settings",
            "Single": output_tmpl % ("truncated",),
        }

        CommandNode.__init__(
            self,
            command=cmd.finalize(),
            threads=threads,
            description="trimming SE adapters from %s"
            % fileutils.describe_files(input_file),
            dependencies=dependencies,
        )


class PE_AdapterRemoval2Node(CommandNode):
    def __init__(
        self,
        input_file_1,
        input_file_2,
        output_prefix,
        collapse=True,
        threads=1,
        options={},
        dependencies=(),
    ):
        # See below for parameters in common between SE/PE
        cmd = AtomicCmdBuilder("AdapterRemoval", CHECK_VERSION=_VERSION_2_CHECK)
        cmd = _set_common_parameters(cmd, threads=threads, options=options)

        # Prefix for output files, to ensure that all end up in temp folder
        cmd.set_option("--basename", "%(TEMP_OUT_BASENAME)s")

        output_tmpl = output_prefix + ".%s.gz"
        cmd.set_kwargs(
            TEMP_OUT_BASENAME=os.path.basename(output_prefix),
            OUT_SETTINGS=output_prefix + ".settings",
            OUT_READS_1=output_tmpl % ("pair1.truncated",),
            OUT_READS_2=output_tmpl % ("pair2.truncated",),
            OUT_SINGLETON=output_tmpl % ("singleton.truncated",),
            OUT_DISCARDED=output_tmpl % ("discarded",),
        )

        self.files = {
            "Statistics": output_prefix + ".settings",
            "Singleton": output_tmpl % ("singleton.truncated",),
            "Paired": output_tmpl % ("pair{Pair}.truncated",),
        }

        if collapse:
            cmd.set_option("--collapse")
            cmd.set_kwargs(
                OUT_COLLAPSED=output_tmpl % ("collapsed",),
                OUT_COLLAPSED_TRUNC=output_tmpl % ("collapsed.truncated",),
            )

            self.files["Collapsed"] = output_tmpl % ("collapsed",)
            self.files["CollapsedTruncated"] = output_tmpl % ("collapsed.truncated",)

        cmd.set_option("--file1", "%(IN_READS_1)s")
        cmd.set_option("--file2", "%(IN_READS_2)s")
        cmd.set_kwargs(IN_READS_1=input_file_1, IN_READS_2=input_file_2)

        apply_options(cmd, options)

        CommandNode.__init__(
            self,
            command=cmd.finalize(),
            threads=threads,
            description="trimming PE adapters from %s"
            % fileutils.describe_paired_files(input_file_1, input_file_2),
            dependencies=dependencies,
        )


class SE_AdapterRemoval3Node(CommandNode):
    def __init__(
        self,
        input_file,
        output_prefix,
        threads=1,
        options=None,
        dependencies=(),
    ):
        options = {} if options is None else dict(options)
        # Option is no longer supported (output is always Phred+33)
        options.pop("--qualitybase-output", None)

        # See below for parameters in common between SE/PE
        cmd = AtomicCmdBuilder("adapterremoval3", CHECK_VERSION=_VERSION_3_CHECK)
        cmd = _set_common_parameters(cmd, threads=threads, options=options)

        # Prefix for output files, ensure that all end up in temp folder
        cmd.set_option("--out-prefix", "%(TEMP_OUT_BASENAME)s")

        output_tmpl = output_prefix + ".%s.fastq.gz"
        cmd.set_kwargs(
            TEMP_OUT_BASENAME=os.path.basename(output_prefix),
            OUT_JSON=output_prefix + ".json",
            OUT_HTML=output_prefix + ".html",
            OUT_MATE_1=output_tmpl % ("r1",),
            OUT_DISCARDED=output_tmpl % ("discarded",),
            IN_READS_1=input_file,
        )

        cmd.set_option("--in-file1", "%(IN_READS_1)s")
        cmd.set_option("--out-discarded", "%(OUT_DISCARDED)s")

        apply_options(cmd, options)

        self.files = {
            "Statistics": output_prefix + ".json",
            "Single": output_tmpl % ("r1",),
        }

        CommandNode.__init__(
            self,
            command=cmd.finalize(),
            threads=threads,
            description="trimming SE adapters from %s"
            % fileutils.describe_files(input_file),
            dependencies=dependencies,
        )


class PE_AdapterRemoval3Node(CommandNode):
    def __init__(
        self,
        input_file_1,
        input_file_2,
        output_prefix,
        collapse=True,
        threads=1,
        options={},
        dependencies=(),
    ):
        options = {} if options is None else dict(options)
        # Option is no longer supported (output is always Phred+33)
        options.pop("--qualitybase-output", None)

        # See below for parameters in common between SE/PE
        cmd = AtomicCmdBuilder("adapterremoval3", CHECK_VERSION=_VERSION_3_CHECK)
        cmd = _set_common_parameters(cmd, threads=threads, options=options)

        # Prefix for output files, to ensure that all end up in temp folder
        cmd.set_option("--out-prefix", "%(TEMP_OUT_BASENAME)s")

        output_tmpl = output_prefix + ".%s.fastq.gz"
        cmd.set_kwargs(
            TEMP_OUT_BASENAME=os.path.basename(output_prefix),
            OUT_JSON=output_prefix + ".json",
            OUT_HTML=output_prefix + ".html",
            OUT_READS_1=output_tmpl % ("r1",),
            OUT_READS_2=output_tmpl % ("r2",),
            OUT_SINGLETON=output_tmpl % ("singleton",),
            OUT_DISCARDED=output_tmpl % ("discarded",),
        )

        self.files = {
            "Statistics": output_prefix + ".json",
            "Singleton": output_tmpl % ("singleton",),
            "Paired": output_tmpl % ("r{Pair}",),
        }

        if collapse:
            cmd.set_option("--merge")
            cmd.set_kwargs(OUT_MERGED=output_tmpl % ("merged",))
            self.files["Collapsed"] = output_tmpl % ("merged",)

        cmd.set_option("--in-file1", "%(IN_READS_1)s")
        cmd.set_option("--in-file2", "%(IN_READS_2)s")
        cmd.set_kwargs(IN_READS_1=input_file_1, IN_READS_2=input_file_2)
        cmd.set_option("--out-discarded", "%(OUT_DISCARDED)s")

        apply_options(cmd, options)

        CommandNode.__init__(
            self,
            command=cmd.finalize(),
            threads=threads,
            description="trimming PE adapters from %s"
            % fileutils.describe_paired_files(input_file_1, input_file_2),
            dependencies=dependencies,
        )


def _set_common_parameters(cmd, options, threads=1):
    # Gzip compress FASTQ files
    cmd.set_option("--gzip")

    # Trim Ns at read ends
    cmd.set_option("--trimns", fixed=False)
    # Trim low quality scores
    cmd.set_option("--trimqualities", fixed=False)

    # Fix number of threads to ensure consistency when scheduling node
    cmd.set_option("--threads", threads)

    # Ensure that any user-specified list of adapters is tracked
    adapter_list = options.pop("--adapter-list", None)
    if adapter_list is not None:
        cmd.set_option("--adapter-list", "%(IN_ADAPTER_LIST)s")
        cmd.set_kwargs(IN_ADAPTER_LIST=adapter_list)

    for key in ("--trim5p", "--trim3p"):
        values = options.pop(key, None)
        if values is not None:
            cmd.add_value(key)
            if isinstance(values, list):
                for value in values:
                    cmd.add_value(value)
            else:
                cmd.add_value(values)

    return cmd



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

from paleomix.common.utilities import safe_coerce_to_tuple
from paleomix.nodes.commands import FilterCollapsedBAMNode
from paleomix.nodes.mapdamage import (
    MapDamageModelNode,
    MapDamagePlotNode,
    MapDamageRescaleNode,
)
from paleomix.nodes.samtools import MarkDupNode
from paleomix.nodes.validation import DetectInputDuplicationNode
from paleomix.pipelines.bam.nodes import index_and_validate_bam


class Library:
    """Represents a single library in a BAM pipeline.

    Is responsible for aggregating per-lane BAMS, removal of PCR duplicates,
    rescaling of quality-scores using mapDamage, as well as running mapDamage
    for QC purposes.

    Properties:
      name      -- Name of the library (as specified in makefile)
      lanes     -- Tuple of lanes assosisated with the library
      options   -- Makefile options that apply to the current library
      folder    -- Folder containing files assosisated with library. Is used as
                   a prefix for files generated by this class.
      bams      -- Dictionary of BAM filenames -> nodes, for each BAM generated by
                   the Library class. Depending on options, this may either be newly
                   generated files, or the files produced by Lanes.
    """

    def __init__(self, config, target, prefix, lanes, name):
        self.name = name
        self.lanes = safe_coerce_to_tuple(lanes)
        self.options = lanes[0].options
        self.folder = os.path.dirname(os.path.dirname(self.lanes[0].folder))

        assert all(
            (self.folder == os.path.dirname(os.path.dirname(lane.folder)))
            for lane in self.lanes
        )
        assert all((self.options == lane.options) for lane in self.lanes)

        lane_bams = self._collect_bams_by_type(self.lanes)

        pcr_duplicates = self.options["Features"]["PCRDuplicates"]
        if pcr_duplicates:
            # pcr_duplicates may be "mark" or any trueish value
            lane_bams = self._remove_pcr_duplicates(
                config, prefix, lane_bams, pcr_duplicates
            )

        # At this point we no longer need to differentiate between read types
        files_and_nodes = self._collect_files_and_nodes(lane_bams)

        # Collect output bams, possible following rescaling
        self.bams, mapdamage_nodes = self._build_mapdamage_nodes(
            config, target, prefix, files_and_nodes
        )

        nodes = [self._build_dataduplication_node(lane_bams)]
        nodes.extend(mapdamage_nodes)

        self.nodes = tuple(nodes)

    @classmethod
    def _collect_bams_by_type(cls, lanes):
        bams = {}
        for lane in lanes:
            for key, files in lane.bams.items():
                key = "collapsed" if (key == "Collapsed") else "normal"
                bams.setdefault(key, {}).update(files)

        return bams

    @classmethod
    def _collect_files_and_nodes(cls, bams):
        files_and_nodes = {}
        for dd in bams.values():
            files_and_nodes.update(dd)
        return files_and_nodes

    def _remove_pcr_duplicates(self, config, prefix, bams, strategy):
        keep_duplicates = isinstance(strategy, str) and strategy.lower() == "mark"

        results = {}
        for key, files_and_nodes in bams.items():
            output_filename = "{}.rmdup.{}.bam".format(self.folder, key)

            if key == "collapsed":
                node = FilterCollapsedBAMNode(
                    input_bams=list(files_and_nodes.keys()),
                    output_bam=output_filename,
                    keep_dupes=keep_duplicates,
                    dependencies=list(files_and_nodes.values()),
                )
            elif key == "normal":
                node = MarkDupNode(
                    in_bams=list(files_and_nodes.keys()),
                    out_bam=output_filename,
                    out_stats="{}.rmdup.{}.stats".format(self.folder, key),
                    keep_duplicates=keep_duplicates,
                    dependencies=list(files_and_nodes.values()),
                )
            else:
                raise RuntimeError("unexpected read type {!r}".format(key))

            # Indexing is required if we wish to calculate per-region statistics
            validated_node = index_and_validate_bam(
                config=config,
                prefix=prefix,
                node=node,
                create_index=bool(prefix.get("RegionsOfInterest")),
            )

            results[key] = {output_filename: validated_node}
        return results

    def _build_mapdamage_nodes(self, config, target, prefix, files_and_nodes):
        # Messing with these does not cause the pipeline to re-do other stuff
        destination = os.path.join(
            config.destination, "%s.%s.mapDamage" % (target, prefix["Name"]), self.name
        )

        run_type = self.options["Features"]["mapDamage"]
        if run_type == "rescale":
            return self._mapdamage_rescale(
                config=config,
                destination=destination,
                prefix=prefix,
                files_and_nodes=files_and_nodes,
            )

        elif run_type == "model":
            # Run of mapDamage including both plots and damage models
            node = self._mapdamage_model(
                destination=destination,
                prefix=prefix,
                files_and_nodes=files_and_nodes,
            )

            return files_and_nodes, (node,)
        elif run_type in ("plot", True):
            # Basic run of mapDamage, only generates plots / tables
            node = self._mapdamage_plot(
                destination=destination,
                prefix=prefix,
                files_and_nodes=files_and_nodes,
            )

            return files_and_nodes, (node,)
        else:
            assert not run_type, run_type
            return files_and_nodes, ()

    def _mapdamage_plot(self, destination, prefix, files_and_nodes):
        title = "mapDamage plot for library %r" % (self.name,)

        return MapDamagePlotNode(
            reference=prefix["Path"],
            input_files=list(files_and_nodes),
            output_directory=destination,
            title=title,
            options=self.options["mapDamage"],
            dependencies=files_and_nodes.values(),
        )

    def _mapdamage_model(self, destination, prefix, files_and_nodes):
        # Generates basic plots / table files
        plot = self._mapdamage_plot(
            destination=destination,
            prefix=prefix,
            files_and_nodes=files_and_nodes,
        )

        # Builds model of postmortem DNA damage
        return MapDamageModelNode(
            reference=prefix["Reference"],
            directory=destination,
            options=self.options["mapDamage"],
            dependencies=plot,
        )

    def _mapdamage_rescale(self, config, destination, prefix, files_and_nodes):
        model = self._mapdamage_model(
            destination=destination,
            prefix=prefix,
            files_and_nodes=files_and_nodes,
        )

        # Rescales BAM quality scores using model built above
        input_files = list(files_and_nodes)
        output_filename = self.folder + ".rescaled.bam"

        scale = MapDamageRescaleNode(
            reference=prefix["Reference"],
            input_files=input_files,
            output_file=output_filename,
            directory=destination,
            options=self.options["mapDamage"],
            dependencies=model,
        )

        # Grab indexing and validation nodes, required by ROIs
        validate = index_and_validate_bam(
            config=config,
            prefix=prefix,
            node=scale,
            create_index=bool(prefix.get("RegionsOfInterest")),
        )

        return {output_filename: validate}, (model,)

    def _build_dataduplication_node(self, bams):
        files_and_nodes = self._collect_files_and_nodes(bams)
        output_file = self.folder + ".duplications_checked"

        return DetectInputDuplicationNode(
            input_files=list(files_and_nodes),
            output_file=output_file,
            dependencies=list(files_and_nodes.values()),
        )



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import copy
import heapq
import itertools


def _safe_coerce(cls):
    def _do_safe_coerce(value):
        if isinstance(value, (str, dict)):
            return cls((value,))

        try:
            return cls(value)
        except TypeError:
            return cls((value,))

    _do_safe_coerce.__doc__ = """Takes a value which be a single object, or an an iterable
        and returns the content wrapped in a {0}. In the case of strings,
        and dictionaries the original string object is returned in a {0},
        and not as a {0} of chars. A TypeError is raised if this is not
        possible (e.g. dict in frozenset).""".format(cls.__name__)
    _do_safe_coerce.__name__ = "safe_coerce_to_{0}".format(cls.__name__)

    return _do_safe_coerce


safe_coerce_to_tuple = _safe_coerce(tuple)
safe_coerce_to_frozenset = _safe_coerce(frozenset)


def try_cast(value, cast_to):
    try:
        return cast_to(value)
    except (ValueError, TypeError):
        return value


def set_in(dictionary, keys, value):
    """Traverses a set of nested dictionaries using the given keys,
    and assigns the specified value to the inner-most
    dictionary (obtained from the second-to-last key), using
    the last key in keys. Thus calling set_in is(d, [X, Y, Z], v)
    is equivalent to calling
      d.setdefault(X, {}).setdefault(Y, {})[Z] = v

    Behavior on non-dictionaries is undefined."""
    keys = list(keys)
    if not keys:
        raise ValueError("No keys passed to 'set_in'!")

    for key in keys[:-1]:
        try:
            dictionary = dictionary[key]
        except KeyError:
            new_dict = {}
            dictionary[key] = new_dict
            dictionary = new_dict

    dictionary[keys[-1]] = value


def get_in(dictionary, keys, default=None):
    """Traverses a set of nested dictionaries using the keys in
    kws, and returns the value assigned to the final keyword
    in the innermost dictionary. Calling get_in(d, [X, Y])
    is equivalent to calling d.get(X).get(Y), with the
    difference that any missing keys causes the default value
    to be returned.

    Behavior on non-dictgionaries is undefined."""
    keys = list(keys)
    for key in keys[:-1]:
        try:
            dictionary = dictionary[key]
        except KeyError:
            return default

    return dictionary.get(keys[-1], default)


def split_before(iterable, pred):
    """Takes a sequence and splits it before every value where pred(v) is true.
    Thus split_before(range(10), key = lambda x: x % 2 == 0) would return the
    sequence [[1], [2,3], [4,5], [6,7], [7,8], [9]]"""
    items = []
    for value in iterable:
        if pred(value) and items:
            yield items
            items = []
        items.append(value)

    if items:
        yield items


# Copied from the Python 'itertools' module documentation
def grouper(size, iterable, fillvalue=None):
    "grouper(3, 'ABCDEFG', 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * size
    return itertools.zip_longest(fillvalue=fillvalue, *args)


def group_by_pred(pred, iterable):
    """Splits items in a sequence into two lists, one containing
    items matching the predicate, and another containing those that
    do not."""
    is_true, is_false = [], []
    for item in iterable:
        if pred(item):
            is_true.append(item)
        else:
            is_false.append(item)

    return is_true, is_false


def fragment(size, lstlike):
    """Faster alternative to grouper for lists/strings."""
    return (lstlike[i : i + size] for i in range(0, len(lstlike), size))


def cumsum(lst, initial=0):
    """Yields the cumulative sums of the values in a
    iterable, starting with the specified initial value."""
    for item in lst:
        initial += item
        yield initial


def fill_dict(destination, source):
    """Returns a copy of 'destination' after setting missing key-
    pairs with copies of those of 'source' recursively."""
    if not isinstance(destination, dict) or not isinstance(source, dict):
        raise TypeError("Non-dictionary parameters in 'fill_dict'")

    def _fill_dict(cur_dest, cur_src):
        for key in cur_src:
            if isinstance(cur_src[key], dict) and isinstance(cur_dest.get(key), dict):
                _fill_dict(cur_dest[key], cur_src[key])
            elif key not in cur_dest:
                cur_dest[key] = cur_src[key]
        return cur_dest

    return _fill_dict(copy.deepcopy(destination), copy.deepcopy(source))


def chain_sorted(*sequences, **kwargs):
    """Chains together sorted sequences, and yields the contents
    in the same order, such that the result is also a sorted sequence.
    The function accepts a 'key'-function keyword, following sort().

    chain_sorted is intended for a few long sequences, and not many short
    sequences. Behavior is undefined if the sequences are not sorted.

    Example:
      >>> tuple(chain_sorted((1, 3, 5), (0, 2, 4)))
      (0, 1, 2, 3, 4, 5)
    """
    key = kwargs.pop("key", None)
    if kwargs:
        raise TypeError(
            "chain_sorted expected keyword 'key', got %r" % (", ".join(kwargs))
        )

    iterators = []
    for index, sequence_iter in enumerate(map(iter, sequences)):
        try:
            current = next(sequence_iter)
            key_value = current if key is None else key(current)

            iterators.append((key_value, index, current, sequence_iter))
        except StopIteration:
            pass

    heapq.heapify(iterators)

    _len, _heappop, _heapreplace = len, heapq.heappop, heapq.heapreplace

    while _len(iterators) > 1:
        last_key_value, index, current, sequence_iter = iterators[0]
        yield current

        for current in sequence_iter:
            key_value = current if key is None else key(current)

            # Optimization for runs of repeated values
            if key_value != last_key_value:
                _heapreplace(iterators, (key_value, index, current, sequence_iter))
                break
            else:
                yield current
        else:
            # No items remaining in top iterator
            _heappop(iterators)

    if _len(iterators) == 1:
        _, _, current, sequence_iter = iterators[0]

        yield current
        for current in sequence_iter:
            yield current


class Immutable:
    """Mixin implementing a immutable class; member variables are specified in
    the init function, cannot be changed afterwards; note that this does not
    prevent changes to the member variables themselves (if not immutable)."""

    def __init__(self, **kwargs):
        object.__init__(self)
        for key, value in kwargs.items():
            object.__setattr__(self, key, value)

    def __setattr__(self, _name, _value):
        raise NotImplementedError("Object is immutable")

    def __delattr__(self, _name):
        raise NotImplementedError("Object is immutable")


class TotallyOrdered:
    """Mixin implementing a rich-comparison interface, provided
    that the subclass implements the less-than operator (__lt__).
    The __lt__ function should return NotImplemented if the other
    object is not the same type.

    The implementation assumes total order:
    http://en.wikipedia.org/wiki/Total_order
    """

    def __lt__(self, other):
        raise NotImplementedError("__lt__ must be implemented!")

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return not ((self < other) or (other < self))

    def __ne__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return not (self == other)

    def __le__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return not (other < self)

    def __ge__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return not (self < other)

    def __gt__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return other < self

    # Shut up warning; if hashable, then the subclass will have
    # to implement the __hash__ member function.
    __hash__ = None



#!/usr/bin/python3
# -*- coding: utf-8 -*-
# Copyright (c) 2016 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import collections
import logging
import os
import re
import tarfile
from io import TextIOWrapper

import pysam

import paleomix.yaml
from paleomix.common.formats.fasta import FASTA
from paleomix.pipelines.zonkey.common import contig_name_to_plink_name, get_sample_names

_SETTINGS_KEYS = (
    "Format",
    "Revision",
    "Plink",
    "NChroms",
    "MitoPadding",
    "SNPDistance",
)


class BAMInfo:
    def __init__(self):
        self.nuclear_contigs = {}
        self.mt_contig = None
        self.mt_length = None
        self.mt_padding = None

    @property
    def is_nuclear(self):
        return bool(self.nuclear_contigs)

    @property
    def is_mitochondrial(self):
        return bool(self.mt_contig)

    def __repr__(self):
        tmpl = "BAMInfo(nuclear=%r, mt_contig=%r, mt_length=%r, mt_padding=%r)"

        return tmpl % (
            self.nuclear_contigs,
            self.mt_contig,
            self.mt_length,
            self.mt_padding,
        )


# Format number for database file; is incremented when the format is changed.
# The 'revision' field specifies updates to the table that do not change the
# format of the database (see below).
_SUPPORTED_DB_FORMAT_MAJOR = 1
_SUPPORTED_DB_FORMAT_MINOR = 20160112

# Required columns in the 'contigs.txt' table; additional columns are ignored
_CONTIGS_TABLE_COLUMNS = frozenset(("ID", "Size", "Checksum"))
# Required columns in the 'samples.txt' table; additional non-group columns are ignored
_SAMPLES_TABLE_COLUMNS = frozenset(("ID", "Species", "Sex", "SampleID", "Publication"))
# Regular expression for parsing Group(K) columns in samples.txt
_SAMPLES_TABLE_GROUP = re.compile(r"^Group\((?P<K>.+)\)$")


class ZonkeyDBError(RuntimeError):
    pass


class ZonkeyDB:
    def __init__(self, filename):
        self.filename = filename

        log = logging.getLogger(__name__)
        log.info("Reading Zonkey database from %r" % (filename,))

        try:
            # Require that the file is not gzip / bzip2 compressed
            _check_file_compression(filename)

            with tarfile.open(filename, "r:") as tar_handle:
                log.info("Reading settings")
                self.settings = self._read_settings(tar_handle, "settings.yaml")
                log.info("Reading list of contigs")
                self.contigs = self._read_contigs_table(tar_handle, "contigs.txt")
                log.info("Reading list of samples")
                self.samples, self.groups = self._read_samples_table(
                    tar_handle, "samples.txt"
                )
                log.info("Reading mitochondrial sequences")
                self.mitochondria = self._read_mitochondria(
                    tar_handle, "mitochondria.fasta"
                )
                log.info("Reading empirical admixture distribution")
                self.simulations = self._read_simulations(tar_handle, "simulations.txt")
                log.info("Determining sample order")
                self.sample_order = self._read_sample_order(tar_handle, "genotypes.txt")
        except (OSError, tarfile.TarError) as error:
            raise ZonkeyDBError(str(error))

        self._cross_validate()

    def validate_bam(self, filename):
        """Validates a sample BAM file, checking that it is either a valid
        mitochondrial BAM (aligned against one of the reference mt sequences),
        or that it is a valid nuclear BAM (aligned against the reference).

        Returns one of INVALID_BAMFILE, NUC_BAMFILE, and MITO_BAMFILE.
        """
        log = logging.getLogger(__name__)
        log.info("Validating BAM file %r ", filename)

        try:
            handle = pysam.AlignmentFile(filename)
        except (ValueError, IOError) as error:
            log.error("Error reading BAM: %s", error)
            return

        return self.validate_bam_handle(handle)

    def validate_bam_handle(self, handle):
        if len(get_sample_names(handle)) > 1:
            log = logging.getLogger(__name__)
            log.warning(
                "BAM read-groups specify more than one sample, "
                "but this tool treats BAMs as a single sample"
            )

        info = BAMInfo()
        if not _validate_mito_bam(self, handle, info):
            return

        if not _validate_nuclear_bam(self, handle, info):
            return

        return info

    def _cross_validate(self):
        """Cross validates tables to ensure consistency."""
        genotypes = set(self.sample_order)
        samples = set(self.samples)
        differences = (genotypes | samples) - (genotypes & samples)
        if differences:
            raise ZonkeyDBError(
                "Mismatch between samples in sample-list and "
                "genotypes table; some samples not found in "
                "both tables: %s" % (",".join(differences),)
            )

        if self.mitochondria is None:
            return

        for name, record in self.mitochondria.items():
            if name not in self.samples:
                # Ignore extra reference sequences
                meta = record.meta.upper()
                if "EXCLUDE" not in list(map(str.strip, meta.split(";"))):
                    raise ZonkeyDBError(
                        "Unexpected mitochondrial sequence: %r" % (name,)
                    )

    @classmethod
    def _read_contigs_table(cls, tar_handle, filename):
        cls._check_required_file(tar_handle, filename)

        table = cls._read_table(tar_handle, filename, _CONTIGS_TABLE_COLUMNS)
        for key, row in table.items():
            try:
                row["Size"] = int(row["Size"])
            except ValueError as error:
                raise ZonkeyDBError(
                    "Invalid size specified for sample %r in "
                    "%r: %r" % (key, filename, error)
                )

            if row["Size"] <= 0:
                raise ZonkeyDBError(
                    "Contig size must be >= 0 for %r in %r, "
                    "not %r" % (key, filename, row["Size"])
                )
        return table

    @classmethod
    def _read_samples_table(cls, tar_handle, filename):
        cls._check_required_file(tar_handle, filename)

        samples = cls._read_table(tar_handle, "samples.txt", _SAMPLES_TABLE_COLUMNS)
        if not samples:
            raise ZonkeyDBError("No samples found in genotypes table!")

        for row in samples.values():
            if row["Sex"].upper() not in ("MALE", "FEMALE", "NA"):
                raise ZonkeyDBError(
                    "Unexpected sample sex (%r); "
                    "expected 'MALE', 'FEMALE', or 'NA'" % (row["Sex"],)
                )

        group_keys = []
        for key in next(iter(samples.values())):
            match = _SAMPLES_TABLE_GROUP.match(key)
            if match is not None:
                k_value = match.groupdict()["K"]
                if not k_value.isdigit():
                    raise ZonkeyDBError(
                        "Malformed Group column name; K is not a number: %r" % (key,)
                    )
                elif not (2 <= int(k_value) <= 7):
                    raise ZonkeyDBError(
                        "K must be between 2 and 7, but found %r" % (key,)
                    )

                group_keys.append((key, int(k_value)))

        groups = {}
        for key, k_value in group_keys:
            group = {}
            for sample_key, sample in samples.items():
                group[sample_key] = sample.pop(key)

            group_labels = frozenset(group.values())
            if group_labels == frozenset("-"):
                continue  # Allowed for backwards compatibility
            elif "-" in group_labels:
                raise ZonkeyDBError(
                    "Not all samples column %r assignd a group" % (key,)
                )
            elif len(group_labels) != k_value:
                raise ZonkeyDBError(
                    "Expected %i groups in column %r, found %i"
                    % (k_value, key, len(group_labels))
                )

            groups[k_value] = group

        if not groups:
            raise ZonkeyDBError("No valid groups in samples.txt")

        return samples, groups

    @classmethod
    def _read_sample_order(cls, tar_handle, filename):
        cls._check_required_file(tar_handle, filename)

        handle = TextIOWrapper(tar_handle.extractfile(filename))
        header = handle.readline().rstrip("\r\n").split("\t")
        sample_order = tuple(header[-1].split(";"))

        if len(sample_order) != len(set(sample_order)):
            raise ZonkeyDBError("Duplicate sample names in %r" % (filename,))

        return sample_order

    def _read_mitochondria(self, tar_handle, filename):
        try:
            tar_handle.getmember(filename)
        except KeyError:
            # Missing MT file is allowed
            return None

        handle = TextIOWrapper(tar_handle.extractfile(filename))

        results = {}
        for record in FASTA.from_lines(handle):
            record = FASTA(
                name=record.name, meta=record.meta, sequence=record.sequence.upper()
            )

            unexpected = set(record.sequence) - set("ACGTN-")
            if unexpected:
                unexpected = ", ".join(map(repr, sorted(unexpected)))
                raise ZonkeyDBError(
                    "Unexpected nucleotide in %s; only A, C, "
                    "G, T, N, and - are allowed, not %s" % (unexpected, filename)
                )
            elif record.name in results:
                raise ZonkeyDBError(
                    "Duplicate sequence name in %s: %r" % (filename, record.name)
                )

            results[record.name] = record

        lengths = frozenset(len(record.sequence) for record in results.values())

        if not lengths:
            raise ZonkeyDBError("No mitochondrial sequences found in %r" % (filename,))
        elif len(lengths) > 2:
            lengths = tuple(sorted(lengths))
            lengths_s = "%s, and %s" % (", ".join(map(str, lengths[:-1])), lengths[-1])

            raise ZonkeyDBError(
                "At most two different sequence lengths "
                "expected for mitochondrial sequences, but "
                "found %i different lengths in %r: %s"
                % (len(lengths), filename, lengths_s)
            )
        elif len(lengths) != 1:
            # Unpadded sequences are allowed
            delta_len = max(lengths) - min(lengths)
            mito_padding = self.settings["MitoPadding"]

            if delta_len != mito_padding:
                raise ZonkeyDBError(
                    "Length difference between mitochondrial "
                    "sequences in %r does not match the "
                    "padding; expected a difference of %i bp, "
                    "but found a %i bp difference."
                    % (filename, mito_padding, delta_len)
                )

        return results

    @classmethod
    def _read_settings(cls, tar_handle, filename):
        cls._check_required_file(tar_handle, filename)

        handle = TextIOWrapper(tar_handle.extractfile(filename))

        try:
            result = paleomix.yaml.safe_load(handle)
        except paleomix.yaml.YAMLError as error:
            raise ZonkeyDBError(
                "Error reading settings file %r; %s" % (filename, error)
            )

        for key in _SETTINGS_KEYS:
            if key != "Plink":
                if not isinstance(result[key], int) or result[key] < 0:
                    raise ZonkeyDBError(
                        "Value for %r in %s must be an non-"
                        "negative integer, not %r" % (key, filename, result[key])
                    )
            elif not isinstance(result[key], str):
                raise ZonkeyDBError(
                    "Value for %r in %s must be a string, "
                    "not %r" % (key, filename, result[key])
                )

        if result["Format"] > _SUPPORTED_DB_FORMAT_MAJOR:
            raise ZonkeyDBError(
                "Database version is too old; this version of "
                "PALEOMIX supports the Zonkey DB v%i, but the "
                "database is v%i; download an updated "
                "database to continue." % (_SUPPORTED_DB_FORMAT_MAJOR, result["Format"])
            )
        elif result["Format"] < _SUPPORTED_DB_FORMAT_MAJOR:
            raise ZonkeyDBError(
                "Database version is too new; this version of "
                "PALEOMIX supports the Zonkey DB v%i, but the "
                "database is v%i; upgrade PALEOMIX to "
                "continue." % (_SUPPORTED_DB_FORMAT_MAJOR, result["Format"])
            )
        elif result["Revision"] < _SUPPORTED_DB_FORMAT_MINOR:
            raise ZonkeyDBError(
                "Database version is too old; this version of "
                "PALEOMIX supports the Zonkey DB v%i, rev. %i "
                "or newer, but the database is v%i rev. %i; "
                "please download an updated database to "
                "continue."
                % (
                    _SUPPORTED_DB_FORMAT_MAJOR,
                    _SUPPORTED_DB_FORMAT_MINOR,
                    result["Format"],
                    result["Revision"],
                )
            )

        return result

    def _read_simulations(self, tar_handle, filename):
        try:
            handle = TextIOWrapper(tar_handle.extractfile(filename))
        except KeyError:
            # Missing simulations file is allowed
            return None

        header = handle.readline().rstrip().split("\t")

        required_keys = set(
            ("NReads", "K", "Sample1", "Sample2", "HasTS", "Percentile", "Value")
        )
        missing_keys = required_keys - set(header)
        if missing_keys:
            missing_keys = ", ".join(map(repr, missing_keys))
            raise ZonkeyDBError(
                "Simulations table %r does not contain all "
                "required columns; columns %r are missing!" % (filename, missing_keys)
            )

        result = []
        for linenum, line in enumerate(handle, start=2):
            fields = line.strip().split("\t")
            if len(fields) != len(header):
                raise ZonkeyDBError(
                    "Line %i in simulations table %r, does "
                    "not contain the expected number of "
                    "columns; expected %i, but found %i!"
                    % (linenum, filename, len(header), len(fields))
                )

            row = dict(zip(header, fields))

            if row["HasTS"] not in ("TRUE", "FALSE"):
                pass

            row["HasTS"] = row["HasTS"] == "TRUE"

            for key in ("NReads", "K"):
                try:
                    row[key] = int(row[key])
                except ValueError:
                    raise ZonkeyDBError(
                        "Malformed value for column %r at "
                        "line %i in simulations table %r; "
                        "expected int, found %r" % (key, linenum, filename, row[key])
                    )

            for key in ("Percentile", "Value"):
                try:
                    row[key] = float(row[key])
                except ValueError:
                    raise ZonkeyDBError(
                        "Malformed value for column %r at "
                        "line %i in simulations table %r; "
                        "expected float, found %r" % (key, linenum, filename, row[key])
                    )

            for key in ("Sample1", "Sample2"):
                groups = frozenset(self.groups[int(row["K"])].values())

                if row[key] not in groups and row[key] != "-":
                    raise ZonkeyDBError(
                        "Invalid group in column %r in "
                        "simulations table %r: %r" % (key, filename, row[key])
                    )

            result.append(row)

        return result

    @classmethod
    def _check_required_file(cls, tar_handle, filename):
        try:
            obj = tar_handle.getmember(filename)
        except KeyError:
            raise ZonkeyDBError(
                "Database does not contain required file %r; "
                "please ensure that this is a valid Zonkey "
                "database file!" % (filename,)
            )

        if not obj.isfile():
            raise ZonkeyDBError(
                "Object %r in Zonkey database is not a "
                "file; please ensure that this is a valid "
                "Zonkey database file!" % (filename,)
            )

    @classmethod
    def _read_table(cls, tar_handle, filename, requied_columns=()):
        requied_columns = frozenset(requied_columns) | frozenset(("ID",))
        handle = TextIOWrapper(tar_handle.extractfile(filename))
        result = {}

        try:
            header = handle.readline().rstrip("\r\n").split("\t")
            if len(header) != len(set(header)):
                raise ZonkeyDBError(
                    "Table %r does contains duplicate columns!" % (filename,)
                )

            if requied_columns - set(header):
                raise ZonkeyDBError(
                    "Required columns are missing in table "
                    "%r: %s" % (filename, ", ".join())
                )

            for linenum, line in enumerate(handle):
                fields = line.rstrip("\r\n").split("\t")

                if len(fields) != len(header):
                    raise ZonkeyDBError(
                        "Error reading %r at line %i; "
                        "expected  %i columns, found %i "
                        "columns!" % (filename, linenum, len(header), len(fields))
                    )

                row = dict(zip(header, fields))
                if row["ID"] in result:
                    raise ZonkeyDBError(
                        "Duplicate IDs in %r: %s" % (filename, row["ID"])
                    )

                result[row["ID"]] = row
        finally:
            handle.close()

        return result


def _validate_mito_bam(data, handle, info):
    if data.mitochondria is None:
        # No mitochondrial data .. skip phylogeny
        return True

    references = handle.references
    min_length = min((len(record.sequence)) for record in data.mitochondria.values())
    log = logging.getLogger(__name__)

    for bam_contig, bam_length in zip(references, handle.lengths):
        if bam_contig not in data.mitochondria:
            continue

        db_sequence = data.mitochondria[bam_contig].sequence
        db_length = len(db_sequence) - db_sequence.count("-")

        if bam_length != db_length:
            log.error(
                "Length of mitochondrial contig %r (%i bp) "
                "does not match the length of the corresponding "
                "sequence in the database (%i bp)" % (bam_contig, bam_length, db_length)
            )
            return False

        filename = handle.filename.decode("utf-8")
        if not os.path.exists(filename + ".bai"):
            log.info("Indexing BAM file %r" % (filename,))
            pysam.index(filename)

        # Workaround for pysam < 0.9 returning list, >= 0.9 returning str
        for line in "".join(pysam.idxstats(filename)).split("\n"):
            line = line.strip()
            if not line:
                continue

            name, _, hits, _ = line.split("\t")
            if (name == bam_contig) and not int(hits):
                log.warning(
                    "Mitochondrial BAM (%r) does not contain "
                    "any reads aligned to contig %r; inferring an "
                    "phylogeny is not possible." % (filename, name)
                )
                return True

        info.mt_contig = bam_contig
        info.mt_length = bam_length
        info.mt_padding = len(db_sequence) - min_length

        return True
    return True


def _validate_nuclear_bam(data, handle, info):
    # Match reference panel contigs with BAM contigs; identification is done
    # by size since different repositories use different naming schemes.
    bam_contigs = collections.defaultdict(list)
    for name, length in zip(handle.references, handle.lengths):
        bam_contigs[length].append(name)

    log = logging.getLogger(__name__)
    panel_names_to_bam = {}
    for name, stats in sorted(data.contigs.items()):
        bam_contig_names = bam_contigs.get(stats["Size"], ())
        if len(bam_contig_names) == 1:
            panel_names_to_bam[name] = bam_contig_names[0]
        elif len(bam_contig_names) > 1:
            candidates = []
            for bam_name in bam_contig_names:
                if contig_name_to_plink_name(bam_name) == name:
                    candidates.append(bam_name)

            if len(candidates) == 1:
                panel_names_to_bam[name] = candidates[0]
            else:
                log.error(
                    "Multiple candidates for chr%s with size %i: %s",
                    name,
                    stats["Size"],
                    ", ".join(bam_contig_names),
                )

    if len(panel_names_to_bam) == len(data.contigs):
        info.nuclear_contigs = panel_names_to_bam
        return True
    elif panel_names_to_bam:
        log.error("Not all nuclear chromosomes found in BAM:")
        for name, stats in sorted(data.contigs.items()):
            is_found = "OK" if name in panel_names_to_bam else "Not found!"
            log.error("  - %s: %s" % (name, is_found))

        return False
    else:
        return True


def _check_file_compression(filename):
    with open(filename, "rb") as handle:
        header = handle.read(2)

    if header == b"\x1f\x8b":
        raise ZonkeyDBError(
            "Zonkey database is gzip compressed; please decompress to continue:\n"
            "  $ gunzip %r" % (filename,)
        )
    elif header == b"BZ":
        raise ZonkeyDBError(
            "Zonkey database is bzip2 compressed; please decompress to continue:\n"
            "  $ bunzip2 %r" % (filename,)
        )



#!/usr/bin/python3
#
# Copyright (c) 2012 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import copy
import logging
import os

import paleomix
import paleomix.common.logging
import paleomix.pipelines.bam.config as bam_config
import paleomix.pipelines.bam.mkfile as bam_mkfile
import paleomix.pipelines.bam.parts as parts
import paleomix.yaml
from paleomix.common import resources
from paleomix.node import NodeError
from paleomix.nodes.bowtie2 import Bowtie2IndexNode
from paleomix.nodes.bwa import BWAIndexNode
from paleomix.nodes.samtools import FastaIndexNode
from paleomix.nodes.validation import ValidateFASTAFilesNode
from paleomix.pipeline import Pypeline
from paleomix.pipelines.bam.makefile import MakefileError, read_makefiles
from paleomix.pipelines.bam.parts import Reads


def build_pipeline_trimming(config, makefile):
    """Builds only the nodes required to produce trimmed reads.
    This reduces the required complexity of the makefile to a minimum."""

    nodes = []
    for _, samples in makefile["Targets"].items():
        for libraries in samples.values():
            for barcodes in libraries.values():
                for record in barcodes.values():
                    if record["Type"] in ("Raw", "Trimmed"):
                        offset = record["Options"]["QualityOffset"]
                        reads = Reads(config, record, offset)

                        nodes.extend(reads.nodes)

    return nodes


def build_pipeline_full(config, makefile, return_nodes=True):
    result = []
    features = makefile["Options"]["Features"]

    if config.validation is None:
        config = copy.copy(config)
        config.validation = features["Validation"]
        if config.validation in (True, False):
            config.validation = "full" if config.validation else "off"

    for target_name, sample_records in makefile["Targets"].items():
        prefixes = []
        for _, prefix in makefile["Prefixes"].items():
            samples = []
            for sample_name, library_records in sample_records.items():
                libraries = []
                for library_name, barcode_records in library_records.items():
                    lanes = []
                    for barcode, record in barcode_records.items():
                        lane = parts.Lane(config, prefix, record, barcode)

                        # ExcludeReads settings may exclude entire lanes
                        if lane.bams:
                            lanes.append(lane)

                    if lanes:
                        libraries.append(
                            parts.Library(
                                config=config,
                                target=target_name,
                                prefix=prefix,
                                lanes=lanes,
                                name=library_name,
                            )
                        )

                if libraries:
                    samples.append(
                        parts.Sample(
                            config=config,
                            prefix=prefix,
                            libraries=libraries,
                            name=sample_name,
                        )
                    )

            if samples:
                prefixes.append(
                    parts.Prefix(
                        config=config,
                        prefix=prefix,
                        samples=samples,
                        features=features,
                        target=target_name,
                    )
                )

        if prefixes:
            target = parts.Target(config, prefixes, target_name)

            # Construct coverage, depth-histogram, and summary nodes, etc.
            parts.add_statistics_nodes(config, makefile, target)

            if return_nodes:
                # Extra tasks (e.g. coverage, depth-histograms, etc.)
                result.extend(target.nodes)
                # Output BAM files
                result.extend(target.bams.values())
            else:
                result.append(target)

    return result


def index_references(config, makefiles):
    references = {}
    references_bwa = {}
    references_bowtie2 = {}
    for makefile in makefiles:
        for subdd in makefile["Prefixes"].values():
            reference = subdd["Reference"]
            if reference not in references:
                # Validation of the FASTA file; not blocking for the other
                # steps, as it is only expected to fail very rarely, but will
                # block subsequent analyses depending on the FASTA.
                valid_node = ValidateFASTAFilesNode(
                    input_files=reference, output_file=reference + ".validated"
                )
                # Indexing of FASTA file using 'samtools faidx'
                faidx_node = FastaIndexNode(reference)

                # Indexing of FASTA file using 'bwa index'
                bwa_node = BWAIndexNode(
                    input_file=reference, dependencies=(valid_node,)
                )
                # Indexing of FASTA file using 'bowtie2-build'
                bowtie2_node = Bowtie2IndexNode(
                    input_file=reference,
                    threads=config.bowtie2_max_threads,
                    index_format=subdd["IdxFmt:Bowtie2"],
                    dependencies=(valid_node,),
                )

                references[reference] = (valid_node, faidx_node)
                references_bwa[reference] = (valid_node, faidx_node, bwa_node)
                references_bowtie2[reference] = (valid_node, faidx_node, bowtie2_node)

            subdd["Nodes"] = references[reference]
            subdd["Nodes:BWA"] = references_bwa[reference]
            subdd["Nodes:Bowtie2"] = references_bowtie2[reference]


def run(config, pipeline_variant):
    paleomix.common.logging.initialize(
        log_level=config.log_level, log_file=config.log_file, name="bam_pipeline"
    )

    logger = logging.getLogger(__name__)
    if pipeline_variant not in ("bam", "trim"):
        logger.critical("Unexpected BAM pipeline variant %r", pipeline_variant)
        return 1

    if not os.path.exists(config.temp_root):
        try:
            os.makedirs(config.temp_root)
        except OSError as error:
            logger.error("Could not create temp root: %s", error)
            return 1

    if not os.access(config.temp_root, os.R_OK | os.W_OK | os.X_OK):
        logger.error("Insufficient permissions for temp root: %r", config.temp_root)
        return 1

    # Init worker-threads before reading in any more data
    pipeline = Pypeline(config)

    try:
        makefiles = read_makefiles(config.makefiles, pipeline_variant)
    except (MakefileError, paleomix.yaml.YAMLError, IOError) as error:
        logger.error("Error reading makefiles: %s", error)
        return 1

    pipeline_func = build_pipeline_trimming
    if pipeline_variant == "bam":
        # Build .fai files for reference .fasta files
        index_references(config, makefiles)

        pipeline_func = build_pipeline_full

    for makefile in makefiles:
        logger.info("Building BAM pipeline for %r", makefile["Filename"])
        try:
            nodes = pipeline_func(config, makefile)
        except (NodeError, MakefileError) as error:
            logger.error("Error while building pipeline for %r", makefile["Filename"])
            logger.error("%s", error)
            return 1

        pipeline.add_nodes(*nodes)

    if config.list_input_files:
        logger.info("Printing output files")
        pipeline.print_input_files()
        return 0
    elif config.list_output_files:
        logger.info("Printing output files")
        pipeline.print_output_files()
        return 0
    elif config.list_executables:
        logger.info("Printing required executables")
        pipeline.print_required_executables()
        return 0

    logger.info("Running BAM pipeline")
    if not pipeline.run(dry_run=config.dry_run, max_threads=config.max_threads):
        return 1

    return 0


def main(argv, pipeline="bam"):
    if pipeline not in ("bam", "trim"):
        raise ValueError(pipeline)

    parser = bam_config.build_parser(pipeline)
    if not argv:
        parser.print_help()
        return 0

    args = parser.parse_args(argv)
    if args.command in ("makefile", "mkfile"):
        return bam_mkfile.main(args, pipeline=pipeline)
    elif args.command in ("example",):
        return resources.copy_example("bam_pipeline", args)

    if args.command.startswith("dry"):
        args.dry_run = True

    return run(args, pipeline_variant=pipeline)



#!/usr/bin/python3
#
# Copyright (c) 2013 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import os

import paleomix
import paleomix.common.logging
from paleomix.common.argparse import SUPPRESS, ArgumentParser
from paleomix.common.procs import max_processes

_DEFAULT_CONFIG_FILES = [
    "/etc/paleomix/phylo_pipeline.ini",
    "~/.paleomix/phylo_pipeline.ini",
]


def build_parser():
    parser = ArgumentParser(
        prog="paleomix phylo_pipeline",
        default_config_files=_DEFAULT_CONFIG_FILES,
    )
    parser.add_argument(
        "commands",
        help="One or more commands separated by '+'. Available commands are 'help', to "
        "display this message; 'example', to create an example project; 'makefile', to "
        "print a makefile template; 'genotype' to perform genotyping on a makefile; "
        "'msa' to perform multiple sequence alignment on a makefile; and 'phylogeny', "
        "to carry out phylogenetic inference on a makefile.",
    )
    parser.add_argument("files", nargs="*", help="One or more YAML files")

    paleomix.common.logging.add_argument_group(parser)

    group = parser.add_argument_group("Scheduling")
    group.add_argument(
        "--examl-max-threads",
        default=1,
        type=int,
        help="Maximum number of threads for each instance of ExaML",
    )
    group.add_argument(
        "--max-threads",
        type=int,
        default=max(2, max_processes()),
        help="Max number of threads to use in total",
    )
    group.add_argument(
        "--dry-run",
        default=False,
        action="store_true",
        help="If passed, only a dry-run in performed, the dependency tree is printed, "
        "and no tasks are executed.",
    )

    group = parser.add_argument_group("Required paths")
    group.add_argument(
        "--temp-root",
        default="./temp",
        type=os.path.abspath,
        help="Location for temporary files and folders",
    )
    group.add_argument(
        "--samples-root",
        default="./data/samples",
        help="Location of BAM files for each sample",
    )
    group.add_argument(
        "--regions-root",
        default="./data/regions",
        help="Location of BED files containing regions of interest",
    )
    group.add_argument(
        "--prefix-root",
        default="./data/prefixes",
        help="Location of prefixes (FASTAs)",
    )
    group.add_argument(
        "--destination",
        default="./results",
        help="The destination folder for result files",
    )

    group = parser.add_argument_group("Files and executables")
    group.add_argument(
        "--list-input-files",
        action="store_true",
        default=False,
        help="List all input files used by pipeline for the "
        "makefile(s), excluding any generated by the "
        "pipeline itself.",
    )
    group.add_argument(
        "--list-output-files",
        action="store_true",
        default=False,
        help="List all output files generated by pipeline for the makefile(s).",
    )
    group.add_argument(
        "--list-executables",
        action="store_true",
        default=False,
        help="List all executables required by the pipeline, "
        "with version requirements (if any).",
    )

    # Removed options
    parser.add_argument("--refseq-root", help=SUPPRESS)
    parser.add_argument("--progress-ui", help=SUPPRESS)
    parser.add_argument("--ui-colors", help=SUPPRESS)

    return parser



#!/usr/bin/python3
# -*- coding: utf-8 -*-
# Copyright (c) 2016 Mikkel Schubert <MikkelSch@gmail.com>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
import collections

import paleomix.yaml

# Format number for database file; is incremented when the format is changed.
# The 'revision' field specifies updates to the table that do not change the
# format of the database (see below).
_SUPPORTED_DB_FORMAT = 1


class DBFileError(RuntimeError):
    pass


def get_sample_names(handle):
    samples = []
    for readgroup in handle.header.get("RG", ()):
        if "SM" in readgroup:
            samples.append(readgroup["SM"])
    return frozenset(samples)


def contig_name_to_plink_name(chrom):
    """Converts chromosome / contig name to the values expected by 'plink',
    namely a digit or X/Y, or returns None if the chromosome could not be
    identified.
    """
    if chrom.isdigit():
        return chrom
    elif chrom.upper() in "XY":
        return chrom.upper()
    elif chrom.lower().startswith("chr") and chrom[3:].isdigit():
        return chrom[3:]
    elif chrom.lower() in ("chrx", "chry"):
        return chrom[3].upper()
    else:
        return None


def read_summary(filename, default="[MISSING VALUE!]"):
    results = collections.defaultdict(lambda: default)
    with open(filename) as handle:
        data = paleomix.yaml.safe_load(handle)

        if not isinstance(data, dict):
            raise DBFileError("Summary file does not contain dictionary")

        results.update(data)

    return results
